#! /usr/bin/env python3

"""
UI Interface controller
"""

from __future__ import annotations

import functools
import textwrap
import typing as ty

from gi.repository import Gdk, Gio, GObject, Gtk

from kupfer import interface
from kupfer.core import actionaccel, settings
from kupfer.core.datactrl import DataController, PaneMode, PaneSel
from kupfer.obj import Action, AnySource, FileLeaf, Leaf
from kupfer.support import pretty, scheduler
from kupfer.ui import (
    accelerators,
    getkey_dialog,
    kupferhelp,
    preferences,
    uievents,
    uiutils,
)
from kupfer.ui._support import escape_markup_str, text_direction_is_ltr
from kupfer.ui.search import ActionSearch, LeafSearch, Search, State

if ty.TYPE_CHECKING:
    from gettext import gettext as _

    from kupfer.core.search import Rankable

_SLOW_INPUT_INTERVAL: ty.Final = 2
_KEY_PRESS_INTERVAL: ty.Final = 0.3
_KEY_PRESS_REPEAT_THRESHOLD: ty.Final = 0.02


# accelerator is any function with no parameters; result is ignored
AccelFunc = ty.Callable[[], ty.Any]


# pylint: disable=too-few-public-methods
@ty.runtime_checkable
class KeyCallback(ty.Protocol):
    """Key press callback interface."""

    def __call__(self, shift_mask: int, mod_mask: int, /) -> bool: ...


_MAX_STRING_LEN: ty.Final[int] = 27


def _trunc_long_str(instr: ty.Any) -> str:
    "truncate long object names"
    ustr = str(instr)
    return (
        ustr[: _MAX_STRING_LEN - 2] + "…"
        if len(ustr) > _MAX_STRING_LEN
        else ustr
    )


def _get_accel(key: str, acf: AccelFunc) -> tuple[str, AccelFunc]:
    """Return name, method pair for @key"""
    if name := accelerators.ACCELERATOR_NAMES.get(key):
        return name, acf

    raise RuntimeError(f"Missing accelerator: {key}")


def _translate_keys(event: Gdk.EventKey) -> tuple[int, int, bool]:
    event_state = event.get_state()
    # translate keys properly
    (
        _was_bound,
        keyv,
        _egroup,
        _level,
        consumed,
    ) = Gdk.Keymap.get_default().translate_keyboard_state(
        event.hardware_keycode, event_state, event.group
    )

    all_modifiers = Gtk.accelerator_get_default_mod_mask()
    shift_mask = (event_state & all_modifiers) == Gdk.ModifierType.SHIFT_MASK
    event_state &= all_modifiers & ~consumed

    return event_state, keyv, shift_mask


# pylint: disable=too-many-public-methods,too-many-instance-attributes
class Interface(GObject.GObject, pretty.OutputMixin):  # type:ignore
    """Controller object that controls the input and the state (current active)
    search object/widget.

    Signals:
    * cancelled: def callback(controller)
        escape was typed


    NOTE: some methods are get by getattr and call!
    """

    __gtype_name__ = "Interface"

    def __init__(self, controller: DataController, window: Gtk.Window) -> None:
        """
        @controller: DataController
        @window: toplevel window
        """
        GObject.GObject.__init__(self)

        self.search = LeafSearch()
        self.search.set_name("kupfer-object-pane")

        self.action = ActionSearch()
        self.action.set_name("kupfer-action-pane")

        self.third = LeafSearch()
        self.third.set_name("kupfer-indirect-object-pane")
        self.current: Search | None = None

        self._widget: Gtk.Widget | None = None
        self._ui_transition_timer = scheduler.Timer()
        self._panel_two_is_visible = True
        self._pane_three_is_visible = False
        self._is_text_mode = False
        self._latest_input_timer = scheduler.Timer()
        self._reset_to_toplevel = False
        self._reset_when_back = False
        self._preedit_text = ""

        self._create_widgets(window)

        self._data_ctrl = controller
        self._data_ctrl.connect("search-result", self._on_search_result)
        self._data_ctrl.connect("source-changed", self._on_source_changed)
        self._data_ctrl.connect("pane-reset", self._on_pane_reset)
        self._data_ctrl.connect("mode-changed", self._on_mode_changed)
        self._data_ctrl.connect(
            "object-stack-changed", self._on_object_stack_changed
        )
        # Setup keyval mapping
        self._prepare_key_book()
        self._action_accel_config = actionaccel.AccelConfig()
        self.search.reset()

    def _create_widgets(self, window: Gtk.Window) -> None:
        setctl = settings.get_settings_controller()

        self._entry = Gtk.Entry()
        self._preedit = Gtk.Entry()
        ## make sure we lose the preedit focus ring
        self._preedit.set_name("kupfer-preedit")
        self._preedit.set_has_frame(False)
        self._preedit.set_width_chars(0)
        self._preedit.set_alignment(1)

        self._label = Gtk.Label()
        self._label.set_width_chars(50)
        self._label.set_max_width_chars(50)
        self._label.set_single_line_mode(True)
        self._label.set_ellipsize(setctl.get_ellipsize_mode())
        self._label.set_name("kupfer-description")

        self._switch_to_source_init()

        self._entry.connect("realize", self._on_entry_realized)
        self._entry.connect("changed", self._on_entry_changed)
        self._preedit.connect("insert-text", self._on_preedit_insert_text)
        self._preedit.connect("draw", self._on_preedit_draw)
        self._preedit.connect("preedit-changed", self._on_preedit_im_changed)
        for widget in (self._entry, self._preedit):
            widget.connect("activate", self._on_activate, None)
            widget.connect("key-press-event", self._on_entry_key_press)
            widget.connect("key-release-event", self._on_entry_key_release)
            widget.connect("copy-clipboard", self._on_entry_copy_clipboard)
            widget.connect("cut-clipboard", self._on_entry_cut_clipboard)
            widget.connect("paste-clipboard", self._on_entry_paste_clipboard)

        # set up panewidget => self signals
        # as well as window => panewidgets
        for widget_owner in (self.search, self.action, self.third):
            widget = widget_owner.widget()
            widget_owner.connect("activate", self._on_activate)
            widget_owner.connect("cursor-changed", self._on_selection_changed)
            widget.connect("button-press-event", self._on_pane_button_press)
            # window signals
            window.connect("configure-event", widget_owner.window_config)
            window.connect("hide", widget_owner.window_hidden)

    def _prepare_key_book(self) -> None:
        keys = (
            "Up",
            "Down",
            "Right",
            "Left",
            "Tab",
            "ISO_Left_Tab",
            "BackSpace",
            "Escape",
            "Delete",
            "space",
            "Page_Up",
            "Page_Down",
            "Home",
            "End",
            "Return",
        )
        self._key_book = key_book = {k: Gdk.keyval_from_name(k) for k in keys}
        if not text_direction_is_ltr():
            # for RTL languages, simply swap the meaning of Left and Right
            # (for keybindings!)
            key_book["Left"], key_book["Right"] = (
                key_book["Right"],
                key_book["Left"],
            )

        self._key_book_cbs: dict[int, KeyCallback] = {
            key_book["Escape"]: self._on_escape_key_press,
            key_book["Up"]: self._on_up_key_press,
            key_book["Page_Up"]: self._on_page_up_key_press,
            key_book["Down"]: self._on_down_key_press,
            key_book["Page_Down"]: self._on_page_down_key_press,
            key_book["Right"]: self._on_right_key_press,
            key_book["BackSpace"]: self._on_backspace_key_press,
            key_book["Left"]: self._on_back_key_press,
            key_book["Tab"]: functools.partial(
                self._on_tab_key_press, reverse=False
            ),
            key_book["ISO_Left_Tab"]: functools.partial(
                self._on_tab_key_press, reverse=True
            ),
            key_book["Home"]: self._on_home_key_press,
        }

    def get_widget(self) -> Gtk.Widget:
        """Return a Widget containing the whole Interface. Create if not exist."""
        if self._widget:
            return self._widget

        box = Gtk.HBox()
        box.pack_start(self.search.widget(), True, True, 3)
        box.pack_start(self.action.widget(), True, True, 3)
        box.pack_start(self.third.widget(), True, True, 3)
        vbox = Gtk.VBox()
        vbox.pack_start(box, True, True, 0)

        label_align = Gtk.Alignment.new(0.5, 1, 0, 0)
        label_align.set_property("top-padding", 3)
        label_align.add(self._label)
        vbox.pack_start(label_align, False, False, 0)
        vbox.pack_start(self._entry, False, False, 0)
        vbox.show_all()
        self.third.hide()
        self._widget = vbox
        return vbox

    def lazy_setup(self) -> None:
        def validate(keystr):
            keyv, mod = Gtk.accelerator_parse(keystr)
            return (
                mod == 0
                and keyv != 0
                and Gtk.accelerator_valid(
                    keyv,
                    Gdk.ModifierType.MOD1_MASK,  # pylint: disable=no-member
                )
            )

        self._action_accel_config.load(validate)
        self.action.action_accel_config = self._action_accel_config
        self.action.lazy_setup()
        self.output_debug("Finished lazy_setup")

    def save_config(self) -> None:
        self._action_accel_config.store()
        self.output_debug("Finished save_config")

    def _on_entry_realized(self, widget: Gtk.Widget) -> None:
        self._update_text_mode()

    def _on_entry_key_release(self, entry, event):
        return

    def _process_accels(self, keyv: int, event_state: int) -> bool:
        setctl = settings.get_settings_controller()
        # process accelerators
        for action, accel in setctl.get_accelerators().items():
            akeyv, amodf = Gtk.accelerator_parse(accel)
            if akeyv and akeyv == keyv and amodf == event_state:
                if action_method := getattr(self, action, None):
                    action_method()
                else:
                    pretty.print_error(__name__, f"Action invalid '{action}'")

                return True

        # look for action accelerators
        if event_state == self.action.accel_modifier:
            keystr = Gtk.accelerator_name(keyv, 0)
            if self._action_accelerator(keystr):
                return True

        return False

    def _on_entry_key_press(  # noqa: PLR0911
        self, entry: Gtk.Entry, event: Gdk.EventKey
    ) -> bool:
        """Intercept arrow keys and manipulate table without losing focus from
        entry field."""
        assert self.current is not None
        direct_text_key: int = Gdk.keyval_from_name("period")
        init_text_keys = list(
            map(Gdk.keyval_from_name, ("slash", "equal", "question"))
        )
        init_text_keys.append(direct_text_key)
        # translate keys properly
        event_state, keyv, shift_mask = _translate_keys(event)

        self._reset_input_timer()
        # process accelerators
        if self._process_accels(keyv, event_state):
            return True

        if self._preedit_text:
            return False

        key_book = self._key_book
        use_command_keys = (
            settings.get_settings_controller().get_use_command_keys()
        )
        has_selection = self.current.get_match_state() == State.MATCH
        if not self._is_text_mode and use_command_keys:
            # translate extra commands to normal commands here
            # and remember skipped chars
            if keyv == key_book["space"]:
                keyv = key_book["Up" if shift_mask else "Down"]

            elif keyv == ord("/") and has_selection:
                keyv = key_book["Right"]

            elif keyv == ord(",") and has_selection:
                if self.comma_trick():
                    return True

            elif keyv in init_text_keys and self._try_enable_text_mode():
                # swallow if it is the direct key
                return keyv == direct_text_key

        if self._is_text_mode and keyv in (
            key_book["Left"],
            key_book["Right"],
            key_book["Home"],
            key_book["End"],
        ):
            # pass these through in text mode
            # except on → at the end of the input
            cursor_position = self._entry.get_property("cursor-position")
            if (
                keyv != key_book["Right"]
                or cursor_position == 0
                or cursor_position != self._entry.get_text_length()
            ):
                return False

        if key_cb := self._key_book_cbs.get(keyv):
            self._reset_to_toplevel = False
            # pylint: disable=no-member
            return key_cb(
                shift_mask, event_state == Gdk.ModifierType.MOD1_MASK
            )

        return False

    def _on_entry_copy_clipboard(self, entry: Gtk.Entry) -> bool:
        """Copy current selection to clipboard. Delegate to text entry when in
        text mode."""
        if self._is_text_mode or not self.current:
            return False

        if selection := self.current.get_current():
            clip = Gtk.Clipboard.get_for_display(
                entry.get_display(), Gdk.SELECTION_CLIPBOARD
            )
            return interface.copy_to_clipboard(selection, clip)

        return False

    def _on_entry_cut_clipboard(self, entry: Gtk.Entry) -> bool:
        if self._on_entry_copy_clipboard(entry):
            self.reset_current()
            self._reset()

        return False

    def _entry_paste_data_received(
        self,
        clipboard: Gtk.Clipboard,
        targets: ty.Iterable[str],
        _extra: ty.Any,
        entry: Gtk.Widget,
    ) -> None:
        uri_target = Gdk.Atom.intern("text/uri-list", False)
        ### check if we can insert files
        if uri_target in targets:
            # paste as files
            sdata = clipboard.wait_for_contents(uri_target)
            self.reset_current()
            self._reset()
            self.put_files(sdata.get_uris(), paths=False)
            ## done
        else:
            # enable text mode and reemit to paste text
            self._try_enable_text_mode()
            if self._is_text_mode:
                entry.emit("paste-clipboard")

    def _on_entry_paste_clipboard(self, entry: Gtk.Widget) -> None:
        if not self._is_text_mode:
            self._reset()
            ## when not in text mode,
            ## stop signal emission so we can handle it
            clipboard = Gtk.Clipboard.get_for_display(
                entry.get_display(), Gdk.SELECTION_CLIPBOARD
            )
            clipboard.request_targets(self._entry_paste_data_received, entry)
            entry.emit_stop_by_name("paste-clipboard")

    def _reset_text(self) -> None:
        """Reset text in entry."""
        self._entry.set_text("")

    def _reset(self) -> None:
        """Reset text and hide table."""
        self._reset_text()
        if self.current:
            self.current.hide_table()

    def reset_current(self, populate: bool = False) -> None:
        """Reset the source or action view.

        Corresponds to backspace.
        """
        assert self.current
        if self.current.get_match_state() is State.WAIT:
            self._toggle_text_mode(False)

        if self.current is self.action or populate:
            self._populate_search()
        else:
            self.current.reset()

    def reset_all(self) -> None:
        """Reset all panes and focus the first.

        This is accelerator handler."""
        self.switch_to_source()
        while self._browse_up():
            pass

        self._toggle_text_mode(False)
        self._data_ctrl.object_stack_clear_all()
        self.reset_current()
        self._reset()

    def _populate_search(self) -> None:
        """Do a blanket search/empty search to populate current pane"""
        pane = self._pane_for_widget(self.current)
        self._data_ctrl.search(pane, interactive=True)

    def soft_reset(self, pane: PaneSel | None = None) -> None:
        """Reset `pane` or current pane context/source softly
        (without visible update), and unset _reset_to_toplevel marker.
        """
        pane = pane or self._pane_for_widget(self.current)
        if newsrc := self._data_ctrl.soft_reset(pane):
            assert self.current
            self.current.set_source(newsrc)

        self._reset_to_toplevel = False

    def _delete_from_stack(self) -> None:
        """Delete item from stack; related do backspace key press."""
        pane = self._pane_for_widget(self.current)
        if self._data_ctrl.get_object_stack(pane):
            self._data_ctrl.object_stack_pop(pane)
            self._reset_text()
            return

        self._on_back_key_press()

    def _relax_search_terms(self) -> None:
        if self._is_text_mode:
            return

        assert self.current
        self._reset_text()
        self.current.relax_match()

    def _get_can_enter_text_mode(self) -> bool:
        """We can enter text mode if the data backend allows,
        and the text entry is ready for input (empty)
        """
        pane = self._pane_for_widget(self.current)
        val = self._data_ctrl.get_can_enter_text_mode(pane)
        entry_text = self._entry.get_text()
        return val and not entry_text

    def _try_enable_text_mode(self) -> bool:
        """Perform a soft reset if possible and then try enabling text mode"""
        if self._reset_to_toplevel:
            self.soft_reset()

        if self._get_can_enter_text_mode():
            return self._toggle_text_mode(True)

        return False

    def _toggle_text_mode(self, val: bool) -> bool:
        """Toggle text mode on/off per @val, and return the subsequent on/off
        state.
        """
        val = val and self._get_can_enter_text_mode()
        self._is_text_mode = val
        self._update_text_mode()
        self._reset()
        return val

    def toggle_text_mode_quick(self) -> None:
        """Toggle text mode or not, if we can or not, without reset.

        NOTE: accelerator
        """
        self._is_text_mode = not self._is_text_mode
        if self._is_text_mode and self.current:
            self.current.hide_table()

        self._update_text_mode()

    def _disable_text_mode_quick(self) -> None:
        """Toggle text mode or not, if we can or not, without reset"""
        if self._is_text_mode:
            self._is_text_mode = False
            self._update_text_mode()

    def _update_text_mode(self) -> None:
        """update appearance to whether text mode enabled or not"""
        if self._is_text_mode:
            self._entry.show()
            self._entry.grab_focus()
            self._entry.set_position(-1)
            self._preedit.hide()
            self._preedit.set_width_chars(0)
        else:
            self._entry.hide()

        self._update_active()

    def _switch_to_source_init(self) -> None:
        """Initial switch to source."""
        self.current = self.search
        self._update_active()
        if self._is_text_mode:
            self.toggle_text_mode_quick()

    def switch_to_source(self) -> None:
        """Switch to first (leaves) pane.

        NOTE: accelerator
        """
        self._switch_current_to(0)

    def switch_to_2(self) -> None:
        """Switch to second (action) pane.

        NOTE: accelerator
        """
        self._switch_current_to(1)

    def switch_to_3(self) -> None:
        """Switch to third (leaves) pane.

        NOTE: accelerator
        """
        self._switch_current_to(2)

    def focus(self) -> None:
        """Called when the interface is focus (after being away)."""
        if self._reset_when_back:
            self._reset_when_back = False
            self._toggle_text_mode(False)

        # preserve text mode, but switch to source if we are not in it
        if not self._is_text_mode:
            self.switch_to_source()

        # Check that items are still valid when "coming back"
        self._data_ctrl.validate()

    def did_launch(self) -> None:
        """Called to notify that 'activate' was successful. Request reset on get
        focus again."""
        self._reset_when_back = True

    def did_get_result(self) -> None:
        """called when a command result has come in."""
        self._reset_when_back = False

    def put_away(self) -> None:
        """Called when the interface is hidden"""
        self._relax_search_terms()
        self._reset_to_toplevel = True
        # no hide / show pane three on put away -> focus anymore

    def select_selected_file(self) -> None:
        """Find & select selected file in browser.

        Note: accelerator.
        """
        # Add optional lookup data to narrow the search
        self._data_ctrl.find_object("qpfer:selectedfile#any.FileLeaf")

    def select_clipboard_file(self) -> None:
        """Find & select leaf representing file in clipboard.

        Note: accelerator.
        """
        # Add optional lookup data to narrow the search
        self._data_ctrl.find_object("qpfer:clipboardfile#any.FileLeaf")

    def select_selected_text(self) -> None:
        """Find & select leaf representing selected text.

        Note: accelerator.
        """
        self._data_ctrl.find_object("qpfer:selectedtext#any.TextLeaf")

    def select_clipboard_text(self) -> None:
        """Find & select leaf representing text in clipboard.

        Note: accelerator.
        """
        # Add optional lookup data to narrow the search
        self._data_ctrl.find_object("qpfer:clipboardtext#any.TextLeaf")

    def select_quit(self) -> None:
        """Find & select quit leaf.

        Note: accelerator.
        """
        self._data_ctrl.reset()
        self._data_ctrl.find_object("qpfer:quit")

    def show_help(self) -> None:
        """Show help.

        Note: accelerator.
        """
        kupferhelp.show_help(self._make_gui_ctx())
        self.emit("launched-action")

    def show_preferences(self) -> None:
        """Show preferences window.

        Note: accelerator.
        """
        self._data_ctrl.reset()
        preferences.show_preferences(self._make_gui_ctx())
        self.emit("launched-action")

    def compose_action(self) -> None:
        """Compose action from current stack.
        NOTE: accelerator
        """
        self._data_ctrl.compose_selection()

    def mark_as_default(self) -> bool:
        """Mark current action as default for selected leaf.

        NOTE: accelerator
        """
        if self.action.get_match_state() != State.MATCH:
            return False

        self._data_ctrl.mark_as_default(PaneSel.ACTION)
        return True

    def erase_affinity_for_first_pane(self) -> bool:
        """
        NOTE: accelerator
        """
        if self.search.get_match_state() != State.MATCH:
            return False

        self._data_ctrl.erase_object_affinity(PaneSel.SOURCE)
        return True

    def comma_trick(self) -> bool:
        """Comma trick - add current leaf to stack.
        NOTE: accelerator
        """
        assert self.current

        if self.current.get_match_state() != State.MATCH:
            return False

        cur = self.current.get_current()
        curpane = self._pane_for_widget(self.current)
        if cur and self._data_ctrl.object_stack_push(curpane, cur):
            self._relax_search_terms()
            if self._is_text_mode:
                self._reset_text()

            return True

        return False

    def _action_accelerator(self, keystr: str) -> bool:
        """Find & launch accelerator action for `keystr` accelerator name..

        Return False if it was not possible to handle or the action was not
        used, return True if it was acted upon
        """
        if self.search.get_match_state() != State.MATCH:
            return False

        self.output_debug("Looking for action accelerator for", keystr)
        success, activate = self.action.select_action_by_accel(keystr)
        if success:
            if activate:
                self._disable_text_mode_quick()
                self.activate()
            else:
                self.switch_to_3()
        else:
            self.output_debug("No action found for", keystr)
            return False

        return True

    def _assign_action_accelerator(self) -> None:
        if self.action.get_match_state() != State.MATCH:
            raise RuntimeError("No Action Selected")

        def is_good_keystr(k: str) -> bool:
            keyv, mods = Gtk.accelerator_parse(k)
            return keyv != 0 and mods in (0, self.action.accel_modifier)

        widget = self.get_widget()
        keystr = getkey_dialog.ask_for_key(
            is_good_keystr,
            screen=widget.get_screen(),
            parent=widget.get_toplevel(),
        )
        if keystr is None:
            # Was cancelled
            return

        action = self.action.get_current()
        # Remove the modifiers
        keyv, _mods = Gtk.accelerator_parse(keystr)
        keystr = Gtk.accelerator_name(keyv, 0)
        assert keystr
        self._action_accel_config.set(action, keystr)

    def get_context_actions(self) -> ty.Iterable[tuple[str, AccelFunc]]:
        """Get a list of (name, function) currently active context actions."""
        assert self.current

        has_match = self.current.get_match_state() == State.MATCH
        if has_match:
            yield _get_accel("compose_action", self.compose_action)

        yield _get_accel("select_selected_text", self.select_selected_text)

        if self._get_can_enter_text_mode():
            yield _get_accel(
                "toggle_text_mode_quick", self.toggle_text_mode_quick
            )

        if self.action.get_match_state() == State.MATCH:
            smatch = self.search.get_current()
            amatch = self.action.get_current()

            label = _('Assign Accelerator to "%(action)s"') % {
                "action": _trunc_long_str(amatch)
            }
            w_label = textwrap.wrap(label, width=40, subsequent_indent="    ")
            yield ("\n".join(w_label), self._assign_action_accelerator)

            label = _('Make "%(action)s" Default for "%(object)s"') % {
                "action": _trunc_long_str(amatch),
                "object": _trunc_long_str(smatch),
            }
            w_label = textwrap.wrap(label, width=40, subsequent_indent="    ")
            yield ("\n".join(w_label), self.mark_as_default)

        if has_match:
            if self._data_ctrl.get_object_has_affinity(PaneSel.SOURCE):
                # TRANS: Removing learned and/or configured bonus search score
                yield (
                    _('Forget About "%s"')
                    % _trunc_long_str(self.search.get_current()),
                    self.erase_affinity_for_first_pane,
                )

            yield _get_accel("reset_all", self.reset_all)

    def _on_pane_reset(
        self,
        _controller: ty.Any,
        pane: int,  # real PaneSel,
        item: Rankable | None,
    ) -> None:
        pane = PaneSel(pane)
        wid = self._widget_for_pane(pane)
        if not item:
            wid.reset()
            return

        wid.set_match_plain(item)
        if wid is self.search:
            self._reset()
            self._toggle_text_mode(False)
            self.switch_to_source()

    def _on_source_changed(
        self,
        _sender: ty.Any,
        pane: int,  # real PaneSel,
        source: AnySource,
        at_root: bool,
        select: ty.Any,
    ) -> None:
        """Notification about a new data source, (represented object for the
        `self.search object`."""
        pane = PaneSel(pane)
        wid = self._widget_for_pane(pane)
        wid.set_source(source)
        wid.reset()
        if pane == PaneSel.SOURCE:
            self.switch_to_source()
            self.action.reset()

        if wid is self.current:
            self._toggle_text_mode(False)
            self._reset_to_toplevel = False
            # when `select` try to show selected item even on root
            if not at_root or select:
                self.reset_current(populate=True)
                wid.show_table_quirk()
                if select:
                    wid.set_match_leaf(select)

    def update_third(self) -> None:
        """Show or hide third panel according to _pane_three_is_visible state."""
        if self._pane_three_is_visible:
            self._ui_transition_timer.set_ms(200, self._show_third_pane, True)
        else:
            self._show_third_pane(False)

    def _on_mode_changed(
        self, _ctr: ty.Any, mode: int, _ignored: ty.Any
    ) -> None:
        """Show / hide third panel on mode changed."""
        if mode == PaneMode.SOURCE_ACTION_OBJECT:
            # use a delay before showing the third pane,
            # but set internal variable to "shown" already now
            self._pane_three_is_visible = True
            self._show_second_pane(True)
            self._ui_transition_timer.set_ms(200, self._show_third_pane, True)
        elif mode == PaneMode.ACTION_OBJECT:
            self._pane_three_is_visible = True
            self._show_second_pane(False)
            self._ui_transition_timer.set_ms(200, self._show_third_pane, True)
        else:
            self._pane_three_is_visible = False
            self._show_second_pane(True)
            self._show_third_pane(False)

    def _show_second_pane(self, show: bool) -> None:
        """Set visibility of second panel to `show`."""
        self._panel_two_is_visible = show
        self.action.set_visible(show)

    def _show_third_pane(self, show: bool) -> None:
        """Set visibility of third panel to `show`."""
        self._ui_transition_timer.invalidate()
        self.third.set_visible(show)

    def _update_active(self) -> None:
        for panewidget in (self.action, self.search, self.third):
            if panewidget is not self.current:
                panewidget.set_state(Gtk.StateType.NORMAL)

            panewidget.match_view.inject_preedit(None)

        assert self.current

        if self._is_text_mode:  # or self._key_repeat_active:
            self.current.set_state(Gtk.StateType.ACTIVE)
        else:
            self.current.set_state(Gtk.StateType.SELECTED)
            self.current.match_view.inject_preedit(self._preedit)

        self._description_changed()

    def _switch_current(self, reverse: bool = False) -> None:
        # Only allow switch if we have match
        if self._pane_three_is_visible:
            curidx = self._pane_for_widget(self.current).value - 1
            newidx = (curidx - 1 if reverse else curidx + 1) % 3
        else:
            # for 2 panels simple switch to other one
            newidx = 0 if self.current == self.action else 1

        self._switch_current_to(newidx)

    def _switch_current_to(self, index: int) -> bool:
        """Switch selected pane.

        index: index (0, 1, or 2) of the pane to select.

        In ACTION_OBJECT mode second panel is hidden.
        """
        order: (
            tuple[LeafSearch, ActionSearch, LeafSearch]
            | tuple[LeafSearch, ActionSearch]
            | tuple[LeafSearch, LeafSearch]
        )

        if not self._panel_two_is_visible:
            order = (self.search, self.third)
        elif self._pane_three_is_visible:
            order = (self.search, self.action, self.third)
        else:
            order = (self.search, self.action)

        if not (0 <= index <= len(order)):
            return False

        pane_before = order[max(index - 1, 0)]
        new_focus = order[index]
        no_match_ok = index == 0
        # Only allow switch if we have match in the pane before
        if (
            self.current
            and (no_match_ok or pane_before.get_match_state() is State.MATCH)
            and new_focus is not self.current
        ):
            self.current.hide_table()
            self.current = new_focus
            # Use toggle_text_mode to reset
            self._toggle_text_mode(False)
            pane = self._pane_for_widget(new_focus)
            self._update_active()
            if self._data_ctrl.get_should_enter_text_mode(pane):
                self.toggle_text_mode_quick()

        return True

    def _browse_up(self) -> bool:
        pane = self._pane_for_widget(self.current)
        return self._data_ctrl.browse_up(pane)

    def _browse_down(self, alternate: bool = False) -> None:
        pane = self._pane_for_widget(self.current)
        self._data_ctrl.browse_down(pane, alternate=alternate)

    def _make_gui_ctx(self) -> uievents.GUIEnvironmentContext:
        event_time = Gtk.get_current_event_time()
        return uievents.gui_context_from_widget(event_time, self._widget)

    def _on_activate(self, _pane_owner: ty.Any, _current: ty.Any) -> None:
        self._data_ctrl.activate(ui_ctx=self._make_gui_ctx())

    def activate(self) -> None:
        """Activate current selection (Run action).

        NOTE: accelerator
        """
        self._on_activate(None, None)

    def execute_file(
        self,
        filepath: str,
        display: str,
        event_time: int,
    ) -> None:
        """Execute a .kfcom file"""

        def _handle_error(exc_info):
            _etype, exc, _tb = exc_info
            if not uiutils.show_notification(str(exc), icon_name="kupfer"):
                raise exc

        ctxenv = uievents.gui_context_from_keyevent(event_time, display)
        self._data_ctrl.execute_file(filepath, ctxenv, _handle_error)

    def _on_search_result(
        self,
        _sender: ty.Any,
        pane: int,  # real PaneSel
        matchrankable: Rankable | None,
        matches: ty.Iterable[Rankable],
        key: str | None,
    ) -> None:
        pane = PaneSel(pane)
        # NOTE: "Always-matching" search.
        # If we receive an empty match, we ignore it, to retain the previous
        # results. The user is not served by being met by empty results.
        if key and len(key) > 1 and matchrankable is None:
            # with typos or so, reset quicker
            self._latest_input_timer.set(
                _SLOW_INPUT_INTERVAL / 2, self._relax_search_terms
            )
            return

        wid = self._widget_for_pane(pane)
        wid.update_match(key, matchrankable, matches)

    def _widget_for_pane(self, pane: PaneSel) -> Search:
        # we have only 3 panels, so this is better than lookup in dict
        if pane == PaneSel.SOURCE:
            return self.search
        if pane == PaneSel.ACTION:
            return self.action
        if pane == PaneSel.OBJECT:
            return self.third

        raise ValueError(f"invalid pane value {pane}")

    def _pane_for_widget(self, widget: GObject.GObject) -> PaneSel:
        # we have only 3 panels, so this is better than lookup in dict
        if widget == self.search:
            return PaneSel.SOURCE
        if widget == self.action:
            return PaneSel.ACTION
        if widget == self.third:
            return PaneSel.OBJECT

        raise ValueError("invalid widget")

    def _on_object_stack_changed(
        self,
        controller: DataController,
        pane: int,  # real PaneSel
    ) -> None:
        """Stack of objects (for comma trick) changed in @pane."""
        pane = PaneSel(pane)
        wid = self._widget_for_pane(pane)
        wid.set_object_stack(controller.get_object_stack(pane))

    def _on_pane_button_press(
        self, widget: Gtk.Widget, event: Gdk.EventButton
    ) -> bool:
        """mouse clicked on a pane widget - activate it."""
        # activate on double-click
        # pylint: disable=no-member,protected-access
        if event.type == Gdk.EventType._2BUTTON_PRESS:  # noqa:SLF001
            self.activate()
            return True

        return False

    def _on_selection_changed(
        self, pane_owner: Search, match: Leaf | Action | None
    ) -> None:
        pane = self._pane_for_widget(pane_owner)
        self._data_ctrl.select(pane, match)
        if pane_owner is not self.current:
            return

        self._description_changed()

    def _description_changed(self) -> None:
        assert self.current
        match = self.current.get_current()
        if desc := match and match.get_description():
            desc = escape_markup_str(desc)
            self._label.set_tooltip_markup(desc)
            self._label.set_has_tooltip(True)

            # show only first line of description
            first_line = desc.split("\n", 1)[0]
            self._label.set_markup(f"<small>{first_line}</small>")

        else:
            # Use invisible WORD JOINER instead of empty, to maintain vertical size
            self._label.set_markup("<small>\N{WORD JOINER}</small>")
            self._label.set_has_tooltip(False)

    def put_text(self, text: str) -> None:
        """Put @text into the interface to search, to use for "queries" from
        other sources."""
        self._try_enable_text_mode()
        self._entry.set_text(text)
        self._entry.set_position(-1)

    def put_files(self, fileuris: ty.Iterable[str], paths: bool) -> None:
        # don't consume iterable
        # self.output_debug("put-files:", list(fileuris))
        if paths:
            objs = (Gio.File.new_for_path(U).get_path() for U in fileuris)
        else:
            objs = (Gio.File.new_for_uri(U).get_path() for U in fileuris)

        if leaves := list(map(FileLeaf, filter(None, objs))):
            self._data_ctrl.insert_objects(
                PaneSel.SOURCE, ty.cast("list[Leaf]", leaves)
            )

    def _reset_input_timer(self) -> None:
        # if input is slow/new, we reset
        self._latest_input_timer.set(
            _SLOW_INPUT_INTERVAL, self._relax_search_terms
        )

    def _on_preedit_im_changed(
        self, _editable: ty.Any, preedit_string: str
    ) -> None:
        """This is called whenever the input method changes its own preedit box.
        We take this opportunity to expand it."""
        if preedit_string:
            assert self.current
            self.current.match_view.expand_preedit(self._preedit)
            self._reset_input_timer()

        self._preedit_text = preedit_string

    def _on_preedit_insert_text(
        self, editable: Gtk.Entry, text: str, byte_length: int, position: int
    ) -> bool:
        # New text about to be inserted in preedit
        if text:
            self._entry.insert_text(text, -1)
            self._entry.set_position(-1)
            self._reset_input_timer()
            self._update_active()

        GObject.signal_stop_emission_by_name(editable, "insert-text")
        return False

    def _on_preedit_draw(self, widget: Gtk.Widget, _cr: ty.Any) -> bool:
        # draw nothing if hidden
        return widget.get_width_chars() == 0  # type: ignore

    def _on_entry_changed(self, editable: Gtk.Entry) -> None:
        """The entry changed callback. if text is blank, start search. Otherwise
        cancel search."""
        text = editable.get_text()

        # draw character count as icon
        editable.set_icon_from_pixbuf(Gtk.EntryIconPosition.SECONDARY, None)

        # cancel search and return if empty
        if not text:
            self._data_ctrl.cancel_search()
            # See if it was a deleting key press
            curr_evt = Gtk.get_current_event()
            if (
                curr_evt
                and curr_evt.type == Gdk.EventType.KEY_PRESS
                and curr_evt.keyval
                in (self._key_book["Delete"], self._key_book["BackSpace"])
            ):
                self._delete_from_stack()

            return

        # start search for updated query
        pane = self._pane_for_widget(self.current)
        if not self._is_text_mode and self._reset_to_toplevel:
            self.soft_reset(pane)

        self._data_ctrl.search(
            pane, key=text, context=text, text_mode=self._is_text_mode
        )

    def _on_up_key_press(self, shift_mask: int, mod_mask: int) -> bool:
        assert self.current
        self.current.go_up()
        return True

    def _on_down_key_press(self, shift_mask: int, mod_mask: int) -> bool:
        assert self.current

        if shift_mask and self.current == self.search:
            self.current.hide_table()
            self._switch_current()

        if (
            not self.current.get_current()
            and self.current.get_match_state() is State.WAIT
        ):
            self._populate_search()

        self.current.go_down()
        return True

    def _on_page_up_key_press(self, shift_mask: int, mod_mask: int) -> bool:
        assert self.current
        self.current.go_page_up()
        return True

    def _on_page_down_key_press(self, shift_mask: int, mod_mask: int) -> bool:
        assert self.current
        if (
            not self.current.get_current()
            and self.current.get_match_state() is State.WAIT
        ):
            self._populate_search()

        self.current.go_page_down()
        return True

    def _on_right_key_press(self, shift_mask: int, mod_mask: int) -> bool:
        # MOD1_MASK is alt/option
        self._browse_down(alternate=bool(mod_mask))
        return True

    def _on_backspace_key_press(self, shift_mask: int, mod_mask: int) -> bool:
        if not self._entry.get_text():  # not has_input
            self._delete_from_stack()
        elif not self._is_text_mode:
            self._entry.delete_text(self._entry.get_text_length() - 1, -1)
        else:
            return False
        return True

    def _on_tab_key_press(
        self, shift_mask: int, mod_mask: int, reverse: bool
    ) -> bool:
        self._switch_current(reverse=reverse)
        return True

    def _on_home_key_press(self, shift_mask: int, mod_mask: int) -> bool:
        assert self.current
        self.current.go_first()
        return True

    def _on_back_key_press(
        self, shift_mask: int = 0, mod_mask: int = 0
    ) -> bool:
        # leftarrow (or backspace without object stack)
        # delete/go up through stource stack
        assert self.current
        if self.current.is_showing_result():
            self.reset_current(populate=True)
        elif not self._browse_up():
            self._reset()
            self.reset_current()
            self._reset_to_toplevel = True

        self._reset_text()
        return True

    def _on_escape_key_press(self, shift_mask: int, mod_mask: int) -> bool:
        """Handle escape if first pane is reset, cancel (put away) self."""
        assert self.current

        if self.current.has_result():
            if self.current.is_showing_result():
                self.reset_current(populate=True)
            else:
                self.reset_current()
        else:
            if self._is_text_mode:
                self._toggle_text_mode(False)
            elif not self.current.get_table_visible():
                pane = self._pane_for_widget(self.current)
                self._data_ctrl.object_stack_clear(pane)
                self.emit("cancelled")

            self._reset_to_toplevel = True
            self.current.hide_table()

        self._reset_text()
        return True

    def re_search_kpfer(self) -> None:
        """Try to re-search current selected KupferObject if it has qf_id.
        For leaves like selected text this update leaf to current selected not
        previously selected.

        If no qf_id is given, reselect current panel.
        """
        if (obj := self.search.get_current()) and (
            qf_id := getattr(obj, "qf_id", None)
        ):
            pretty.print_debug(__name__, f"re-search qpfer '{qf_id}'")
            self._data_ctrl.find_object(f"qpfer:{qf_id}")
        else:
            pane = self._pane_for_widget(self.current)
            self._data_ctrl.reselect(pane)

    def find_object(self, qpfer: str) -> None:
        self._data_ctrl.find_object(qpfer)


GObject.type_register(Interface)
GObject.signal_new(
    "cancelled",
    Interface,
    GObject.SignalFlags.RUN_LAST,
    GObject.TYPE_BOOLEAN,
    (),
)
# Send only when the interface itself launched an action directly
GObject.signal_new(
    "launched-action",
    Interface,
    GObject.SignalFlags.RUN_LAST,
    GObject.TYPE_BOOLEAN,
    (),
)



from __future__ import annotations

__kupfer_name__ = _("Search the Web")
__kupfer_sources__ = ("OpenSearchSource",)
__kupfer_text_sources__ = ()
__kupfer_actions__ = (
    "SearchFor",
    "SearchWithEngine",
    "SearchWithDefaultEngine",
)
__description__ = _(
    "Search the web with OpenSearch and user defined search engines"
)
__version__ = "2023-12-23"
__author__ = "Ulrik Sverdrup <ulrik.sverdrup@gmail.com>, KB"

import locale
import os
import typing as ty
import urllib.parse
from pathlib import Path
from xml.etree import ElementTree as ET

from kupfer import config, launch, plugin_support
from kupfer.obj import Action, Leaf, Source, TextLeaf
from kupfer.plugin._firefox_support import get_firefox_home_file

if ty.TYPE_CHECKING:
    from gettext import gettext as _

# default search engine when user no configure default_engine and there is no
# extra_engines
_DEFAULT_ENGINE = "https://duckduckgo.com/?q=%s"

__kupfer_settings__ = plugin_support.PluginSettings(
    {
        "key": "extra_engines",
        "label": _("User search engines:"),
        "type": str,
        "multiline": True,
        "value": "https://www.qwant.com/?q=%s\n"
        "https://search.brave.com/search?q=%s",
        "tooltip": _(
            "Define URLs for search engines; '%s' is replaced by search term."
        ),
    },
    {
        "key": "default_engine",
        "label": _("Default search engine:"),
        "type": str,
        "value": "",
        "tooltip": _(
            "Define URL for default search engine; '%s' is replaced by "
            "search term. If empty, use first engine from 'User search engines'"
            " or DuckDuckGo"
        ),
    },
)


def _noescape_urlencode(items):
    """Assemble an url param string from @items, without
    using any url encoding.
    """
    return "?" + "&".join(f"{n}={v}" for n, v in items)


def _urlencode(word):
    """Urlencode a single string of bytes @word"""
    return urllib.parse.urlencode({"q": word})[2:]


def _do_search_engine(terms, search_url, encoding="UTF-8"):
    """Show an url searching for @search_url with @terms"""
    terms = _urlencode(terms)
    if "{searchTerms}" in search_url:
        query_url = search_url.replace("{searchTerms}", terms)
    else:
        query_url = search_url.replace("%s", terms)

    launch.show_url(query_url)


def _get_custom_engine_name(url: str) -> str | None:
    if not url:
        return None

    components = urllib.parse.urlparse(url)
    if netloc := components.netloc:
        return (
            netloc.removeprefix("www.")
            .removesuffix(".com")
            .replace(".", " ")
            .capitalize()
        )

    return None


class SearchWithDefaultEngine(Action):
    """Search for term in default search engine."""

    action_accelerator = "s"

    def __init__(self):
        Action.__init__(self, _("Search..."))

    def activate(self, leaf, iobj=None, ctx=None):
        url = __kupfer_settings__["default_engine"].strip()
        if not url and (urls := __kupfer_settings__["extra_engines"]):
            url = urls.replace(";", "\n").split("\n", 1)[0].strip()

        url = url or _DEFAULT_ENGINE
        _do_search_engine(leaf.object, url)

    def item_types(self):
        yield TextLeaf

    def get_description(self):
        return _("Search the web with default search engines")

    def get_icon_name(self):
        return "edit-find"


class SearchWithEngine(Action):
    """TextLeaf -> SearchWithEngine -> SearchEngine"""

    def __init__(self):
        Action.__init__(self, _("Search With..."))

    def activate(self, leaf, iobj=None, ctx=None):
        coding = iobj.object.get("InputEncoding")
        url = iobj.object["Url"]
        _do_search_engine(leaf.object, url, encoding=coding)

    def item_types(self):
        yield TextLeaf

    def requires_object(self):
        return True

    def object_types(self):
        yield SearchEngine

    def object_source(self, for_item=None):
        return OpenSearchSource()

    def get_description(self):
        return _("Search the web with OpenSearch search engines")

    def get_icon_name(self):
        return "edit-find"


class SearchFor(Action):
    """SearchEngine -> SearchFor -> TextLeaf

    This is the opposite action to SearchWithEngine
    """

    action_accelerator = "s"

    def __init__(self):
        Action.__init__(self, _("Search For..."))

    def activate(self, leaf, iobj=None, ctx=None):
        coding = leaf.object.get("InputEncoding")
        url = leaf.object["Url"]
        terms = iobj.object
        _do_search_engine(terms, url, encoding=coding)

    def item_types(self):
        yield SearchEngine

    def requires_object(self):
        return True

    def object_types(self):
        yield TextLeaf

    def get_description(self):
        return _("Search the web with OpenSearch search engines")

    def get_icon_name(self):
        return "edit-find"


class SearchEngine(Leaf):
    def get_description(self):
        desc = self.object.get("Description")
        return desc if desc != str(self) else None

    def get_icon_name(self):
        return "text-html"


class OpenSearchParseError(Exception):
    pass


def gettagname(tag):
    return tag.rsplit("}", 1)[-1]


def _get_plugin_dirs() -> ty.Iterator[Path]:
    """Get all possible plugins path (may not exists)"""
    # accept in kupfer data dirs
    yield from map(Path, config.get_data_dirs("searchplugins"))

    # firefox in home directory
    if ffx_home := get_firefox_home_file("searchplugins"):
        yield ffx_home

    yield from map(
        Path, config.get_data_dirs("searchplugins", package="firefox")
    )
    yield from map(
        Path, config.get_data_dirs("searchplugins", package="iceweasel")
    )

    suffixes = ["en-US"]
    if cur_lang := locale.getlocale(locale.LC_MESSAGES)[0]:
        suffixes = [cur_lang.replace("_", "-"), cur_lang[:2], *suffixes]

    addon_dir = Path("/usr/lib/firefox-addons/searchplugins")
    for suffix in suffixes:
        if (addon_lang_dir := addon_dir.joinpath(suffix)).exists():
            yield addon_lang_dir
            break

    # debian iceweasel
    yield Path("/etc/iceweasel/searchplugins/common")

    for suffix in suffixes:
        yield Path("/etc/iceweasel/searchplugins/locale", suffix)

    # try to find all versions of firefox
    for prefix in ("/usr/lib", "/usr/share"):
        with os.scandir(prefix) as entries:
            for entry in entries:
                if entry.name.startswith(("firefox", "iceweasel")):
                    yield Path(entry.path, "searchplugins")
                    yield Path(
                        entry.path, "distribution", "searchplugins", "common"
                    )


_OS_VITAL_KEYS = {"Url", "ShortName"}
_OS_KEYS = ("Description", "Url", "ShortName", "InputEncoding")
_OS_ROOTS = ("OpenSearchDescription", "SearchPlugin")


def _parse_etree(etree, name=None):
    if gettagname(etree.getroot().tag) not in _OS_ROOTS:
        raise OpenSearchParseError(f"Search {name} has wrong type")

    search = {}
    for child in etree.getroot():
        tagname = gettagname(child.tag)
        if tagname not in _OS_KEYS:
            continue

        # Only pick up Url tags with type="text/html"
        if tagname == "Url":
            if child.get("type") == "text/html" and child.get("template"):
                text = child.get("template")
                params = tuple(
                    (ch.get("name"), ch.get("value"))
                    for ch in child
                    if gettagname(ch.tag) == "Param"
                )
                if params:
                    text += _noescape_urlencode(params)
            else:
                continue

        else:
            text = (child.text or "").strip()

        search[tagname] = text

    if not _OS_VITAL_KEYS.issubset(list(search.keys())):
        raise OpenSearchParseError(f"Search {name} missing _OS_KEYS")

    return search


class OpenSearchSource(Source):
    def __init__(self):
        Source.__init__(self, _("Search Engines"))

    def initialize(self):
        __kupfer_settings__.connect(
            "plugin-setting-changed", self._on_setting_changed
        )

    def _on_setting_changed(self, settings, key, value):
        self.mark_for_update()

    def _parse_opensearch(self, path: str) -> dict[str, ty.Any] | None:
        try:
            etree = ET.parse(path)
            return _parse_etree(etree, name=path)  # type:ignore
        except Exception as exc:
            self.output_debug(f"{type(exc).__name__}: {exc}")

        return None

    def get_items(self) -> ty.Iterator[SearchEngine]:
        # files are unique by filename to allow override
        visited_files = set()
        for pdir in _get_plugin_dirs():
            if not pdir.is_dir():
                continue

            self.output_debug("Processing searchplugins dir", pdir)

            with os.scandir(pdir) as entries:
                for entry in entries:
                    if entry.path in visited_files:
                        continue

                    visited_files.add(entry.path)
                    if not entry.is_dir() and (
                        search := self._parse_opensearch(entry.path)
                    ):
                        yield SearchEngine(search, search["ShortName"])

        # add user search engines
        if custom_ses := __kupfer_settings__["extra_engines"]:
            for url in custom_ses.replace(";", "\n").split():
                if name := _get_custom_engine_name(url.strip()):
                    yield SearchEngine({"Url": url, "Description": url}, name)

    def should_sort_lexically(self):
        return True

    def provides(self):
        yield SearchEngine

    def get_icon_name(self):
        return "applications-internet"



"""
It *should* be possible to support Tomboy and Gnote equally since
they support the same DBus protocol. This plugin takes this assumption.
"""

__kupfer_name__ = _("Notes")
__kupfer_sources__ = ("NotesSource",)
__kupfer_actions__ = (
    "AppendToNote",
    "AppendTextToNote",
    "CreateNote",
    "GetNoteSearchResults",
)
__description__ = _("Gnote or Tomboy notes")
__version__ = "2017.2"
__author__ = ""

import os
import time
import typing as ty
import xml.sax.saxutils

import dbus
from gi.repository import GLib
from xdg import BaseDirectory

from kupfer import icons, plugin_support
from kupfer.obj import (
    Action,
    Leaf,
    NotAvailableError,
    Source,
    TextLeaf,
    TextSource,
)
from kupfer.obj.apps import ApplicationSource
from kupfer.support import pretty, textutils, weaklib

if ty.TYPE_CHECKING:
    from gettext import gettext as _


PROGRAM_IDS = ["gnote", "tomboy", "kzrnote"]
__kupfer_settings__ = plugin_support.PluginSettings(
    {
        "key": "notes_application",
        "label": _("Work with application"),
        "type": str,
        "value": "",
        "alternatives": ["", *PROGRAM_IDS],
    },
)

plugin_support.check_dbus_connection()


## Tuples of  service name, object name, interface name
_PROGRAM_SERVICES = {
    "gnote": (
        "org.gnome.Gnote",
        "/org/gnome/Gnote/RemoteControl",
        "org.gnome.Gnote.RemoteControl",
    ),
    "tomboy": (
        "org.gnome.Tomboy",
        "/org/gnome/Tomboy/RemoteControl",
        "org.gnome.Tomboy.RemoteControl",
    ),
    "kzrnote": (
        "io.github.kupferlauncher.kzrnote",
        "/io/github/kupferlauncher/kzrnote",
        "io.github.kupferlauncher.kzrnote",
    ),
}


def _get_notes_interface(activate=False):
    """Return the dbus proxy object for our Note Application.

    if @activate, we will activate it over d-bus (start if not running)
    """
    bus = dbus.SessionBus()

    set_prog = __kupfer_settings__["notes_application"]
    programs = (set_prog,) if set_prog else PROGRAM_IDS

    for program in programs:
        service_name, obj_name, iface_name = _PROGRAM_SERVICES[program]
        if activate:
            bus.start_service_by_name(service_name)
        elif not bus.name_has_owner(service_name):
            continue

        try:
            searchobj = bus.get_object(service_name, obj_name)
        except dbus.DBusException as exc:
            pretty.print_error(__name__, exc)
            return None

        return dbus.Interface(searchobj, iface_name)

    return None


def _get_notes_interactive():
    """Return the dbus proxy object, activate if necessary,
    raise an OperationError if not available."""
    if (obj := _get_notes_interface(activate=True)) is not None:
        return obj

    raise NotAvailableError(__kupfer_settings__["notes_application"])


def _reply_noop(*args):
    pass


class RetryDbusCalls(pretty.OutputMixin):
    """A d-bus interface wrapper for a proxy object; will retry a method
    call if it fails (a limited number of times).

    The method call must be async (with reply_handler and error_handler)
    """

    def __init__(self, proxy_object, retries=10):
        self.__obj = proxy_object
        self.__retries = retries

    @property
    def proxy_obj(self):
        """Return the inner proxy object. You can call methods synchronously on
        it."""
        return self.__obj

    def __getattr__(self, name):
        x = 0

        def proxy_method(*args, error_handler=None, **kwargs):
            def make_call():
                getattr(self.__obj, name)(
                    *args, error_handler=error_handler_, **kwargs
                )

            def error_handler_(exc):
                nonlocal x
                x += 1
                if (
                    x > self.__retries
                    or exc.get_dbus_name()
                    != "org.freedesktop.DBus.Error.UnknownMethod"
                ):
                    return error_handler(exc)

                self.output_debug("retrying", name, "because of", exc)
                GLib.timeout_add(25 * x, make_call)
                return None

            return make_call()

        return proxy_method


def _make_error_handler(ctx):
    def error_handler(exc):
        pretty.print_debug(__name__, exc)
        ctx.register_late_error(
            NotAvailableError(__kupfer_settings__["notes_application"])
        )

    return error_handler


class Open(Action):
    action_accelerator = "o"

    def __init__(self):
        Action.__init__(self, _("Open"))

    def wants_context(self):
        return True

    def activate(self, leaf, iobj=None, ctx=None):
        assert ctx

        noteuri = leaf.object
        notes = RetryDbusCalls(_get_notes_interactive())
        notes.DisplayNote(
            noteuri,
            reply_handler=_reply_noop,
            error_handler=_make_error_handler(ctx),
        )

    def get_description(self):
        return _("Open with notes application")

    def get_gicon(self):
        app_icon = icons.get_gicon_with_fallbacks(None, PROGRAM_IDS)
        return icons.ComposedIcon(self.get_icon_name(), app_icon)


class AppendToNote(Action):
    def __init__(self, name=None):
        Action.__init__(self, name or _("Append to Note..."))

    def wants_context(self):
        return True

    def activate(self, leaf, iobj=None, ctx=None):
        assert iobj
        assert ctx

        notes = RetryDbusCalls(_get_notes_interactive())
        noteuri = iobj.object
        text = leaf.object

        def reply_note_plain_text(contents):
            if not contents.endswith("\n"):
                contents += "\n"

            contents += text
            if not text.endswith("\n"):
                contents += "\n"

            notes.SetNoteContents(
                noteuri,
                contents,
                reply_handler=_reply_noop,
                error_handler=_make_error_handler(ctx),
            )

        def reply_note_xml(xmlcontents):
            # NOTE: We search and replace in the XML here
            endtag = "</note-content>"
            xmltext = xml.sax.saxutils.escape(text)
            xmlcontents = xmlcontents.replace(endtag, f"\n{xmltext}{endtag}")
            notes.SetNoteCompleteXml(
                noteuri,
                xmlcontents,
                reply_handler=_reply_noop,
                error_handler=_make_error_handler(ctx),
            )

        if __kupfer_settings__["notes_application"] == "kzrnote":
            notes.GetNoteContents(
                noteuri,
                reply_handler=reply_note_plain_text,
                error_handler=_make_error_handler(ctx),
            )
        else:
            notes.GetNoteCompleteXml(
                noteuri,
                reply_handler=reply_note_xml,
                error_handler=_make_error_handler(ctx),
            )

    def item_types(self):
        yield TextLeaf

    def requires_object(self):
        return True

    def object_types(self):
        yield Note

    def object_source(self, for_item=None):
        return NotesSource()

    def get_description(self):
        return _("Add text to existing note")

    def get_icon_name(self):
        return "list-add"


class AppendTextToNote(AppendToNote):
    def __init__(self):
        super().__init__(_("Append..."))

    def wants_context(self):
        return True

    def activate(self, leaf, iobj=None, ctx=None):
        assert iobj
        leaf, iobj = iobj, leaf
        return super().activate(leaf, iobj, ctx)

    def item_types(self):
        return super().object_types()

    def requires_object(self):
        return True

    def object_types(self):
        return super().item_types()

    def object_source(self, for_item=None):
        return TextSource()


def _prepare_note_text(text):
    ## split the text into a title + newline + rest of the text
    ## if we only get the title, put in two helpful newlines
    title, body = textutils.extract_title_body(text)
    if body.lstrip():
        return f"{title}\n{body}"

    return f"{title}\n\n"


class CreateNote(Action):
    def __init__(self):
        Action.__init__(self, _("Create Note"))

    def wants_context(self):
        return True

    def activate(self, leaf, iobj=None, ctx=None):
        assert ctx

        notes = RetryDbusCalls(_get_notes_interactive())
        text = _prepare_note_text(leaf.object)

        def _created_note(noteuri):
            nonlocal notes
            notes = notes.proxy_obj
            # FIXME: For Gnote we have to call DisplayNote
            # else we can't change its contents
            notes.DisplayNote(noteuri)
            notes.SetNoteContents(noteuri, text)

        notes.CreateNote(
            reply_handler=_created_note, error_handler=_make_error_handler(ctx)
        )

    def item_types(self):
        yield TextLeaf

    def get_description(self):
        return _("Create a new note from this text")

    def get_icon_name(self):
        return "document-new"


class GetNoteSearchResults(Action):
    def __init__(self):
        Action.__init__(self, _("Get Note Search Results..."))

    def is_factory(self):
        return True

    def wants_context(self):
        return True

    def activate(self, leaf, iobj=None, ctx=None):
        assert ctx

        query = leaf.object.lower()
        notes = RetryDbusCalls(_get_notes_interactive())

        def search_reply(noteuris):
            ctx.register_late_result(NoteSearchSource(query, noteuris))

        notes.SearchNotes(
            query,
            False,
            reply_handler=search_reply,
            error_handler=_make_error_handler(ctx),
        )

    def item_types(self):
        yield TextLeaf

    def get_description(self):
        return _("Show search results for this query")


class NoteSearchSource(Source):
    def __init__(self, query, noteuris):
        self.query = query.lower()
        Source.__init__(self, _("Notes"))
        self.noteuris = noteuris

    def get_items(self):
        notes = _get_notes_interactive()
        for noteuri in self.noteuris:
            title = notes.GetNoteTitle(noteuri)
            date = notes.GetNoteChangeDate(noteuri)
            yield Note(noteuri, title, date)

    def repr_key(self):
        return self.query

    def get_gicon(self):
        return icons.get_gicon_with_fallbacks(None, PROGRAM_IDS)

    def provides(self):
        yield Note


class Note(Leaf):
    """The Note Leaf's represented object is the Note URI"""

    def __init__(self, obj, name, date):
        self.changedate = date
        Leaf.__init__(self, obj, name)

    def get_actions(self):
        yield Open()

    def repr_key(self):
        # the Note URI is unique&persistent for each note
        return self.object

    def get_description(self):
        today_date = time.localtime()[:3]
        yest_date = time.localtime(time.time() - 3600 * 24)[:3]
        change_time = time.localtime(self.changedate)

        if today_date == change_time[:3]:
            time_str = _("today, %s") % time.strftime("%X", change_time)
        elif yest_date == change_time[:3]:
            time_str = _("yesterday, %s") % time.strftime("%X", change_time)
        else:
            time_str = time.strftime("%c", change_time)
        # TRANS: Note description, %s is last changed time in locale format
        return _("Last updated %s") % time_str

    def get_icon_name(self):
        return "text-x-generic"


class ClassProperty(property):
    """Subclass property to make classmethod properties possible"""

    def __get__(
        self, cls: ty.Any, owner: ty.Optional[type] = None, /
    ) -> ty.Any:
        # pylint: disable=no-member
        return self.fget.__get__(None, owner)()  # type: ignore


class NotesSource(ApplicationSource):
    source_scan_interval: int = 3600

    def __init__(self):
        super().__init__(_("Notes"))
        self._notes = []
        self.monitor_token = None

    def initialize(self):
        """Set up filesystem monitors to catch changes"""
        # We monitor all directories that exist of a couple of candidates
        dirs: list[str] = []
        for program in PROGRAM_IDS:
            dirs.extend(
                (
                    os.path.join(BaseDirectory.xdg_data_home, program),
                    os.path.expanduser(f"~/.{program}"),
                )
            )

        self.monitor_token = self.monitor_directories(*dirs)

        set_prog = __kupfer_settings__["notes_application"]
        if set_prog in _PROGRAM_SERVICES:
            bus_name = _PROGRAM_SERVICES[set_prog][0]
            bus = dbus.SessionBus()
            weaklib.dbus_signal_connect_weakly(
                bus,
                "NameOwnerChanged",
                self._name_owner_changed,
                dbus_interface="org.freedesktop.DBus",
                arg0=bus_name,
            )

    def _name_owner_changed(self, name, old, new):
        if new:
            self.mark_for_update()

    def _update_cache(self, notes):
        try:
            noteuris = notes.ListAllNotes()
        except dbus.DBusException as exc:
            self.output_error(f"{type(exc).__name__}: {exc}")
            return

        templates = notes.GetAllNotesWithTag("system:template")

        self._notes = []
        for noteuri in noteuris:
            if noteuri not in templates:
                title = notes.GetNoteTitle(noteuri)
                date = notes.GetNoteChangeDate(noteuri)
                self._notes.append((noteuri, title, date))

    def get_items(self):
        notes = _get_notes_interface()
        if notes:
            self._update_cache(notes)

        for noteuri, title, date in self._notes:
            yield Note(noteuri, title, date=date)

    def provides(self):
        yield Note

    def get_gicon(self):
        return icons.get_gicon_with_fallbacks(None, PROGRAM_IDS)

    def get_icon_name(self):
        return "gnote"

    @ClassProperty
    @classmethod
    def appleaf_content_id(cls):
        return __kupfer_settings__["notes_application"] or PROGRAM_IDS



"""
Object related to composited objects.

This file is a part of the program kupfer, which is
released under GNU General Public License v3 (or any later version),
see the main program file, and COPYING for details.
"""

from __future__ import annotations

import typing as ty

from kupfer import icons, puid
from kupfer.obj import actions, exceptions, objects
from kupfer.obj.base import Action, Leaf, Source
from kupfer.support import itertools, pretty, scheduler, textutils

if ty.TYPE_CHECKING:
    from gettext import gettext as _
    from gettext import ngettext

    from kupfer.core import commandexec


__all__ = (
    "ComposedLeaf",
    "MultipleLeaf",
    "TimedPerform",
)


class TimedPerform(actions.Perform):
    """A timed (delayed) version of Run (Perform)"""

    action_accelerator: str | None = None

    def __init__(self, name: str = _("Run after Delay...")) -> None:
        super().__init__(name)

    def activate(
        self,
        leaf: Leaf,
        iobj: Leaf | None = None,
        ctx: commandexec.ExecutionToken | None = None,
    ) -> Leaf | None:
        assert isinstance(leaf, objects.RunnableLeaf)
        assert isinstance(iobj, objects.TextLeaf)
        # make a timer that will fire when Kupfer exits
        interval = textutils.parse_time_interval(iobj.object)
        pretty.print_debug(__name__, f"Run {leaf} in {interval} seconds")
        timer = scheduler.Timer(True)
        if leaf.wants_context():
            timer.set(interval, leaf.run, ctx)
        else:
            timer.set(interval, leaf.run)

        return None

    def requires_object(self) -> bool:
        return True

    def object_types(self) -> ty.Iterator[ty.Type[Leaf]]:
        yield objects.TextLeaf

    def valid_object(self, iobj: Leaf, for_item: Leaf | None = None) -> bool:
        return textutils.parse_time_interval(iobj.object) > 0

    def get_description(self) -> str:
        return _("Perform command after a specified time interval")


class ComposedLeaf(objects.RunnableLeaf):
    """Leaf that contains many other leaves, created by compose action.
    Represented object is tuple (object, action, action object).
    """

    serializable = 1

    def __init__(
        self, obj: Leaf, action: Action, iobj: Leaf | None = None
    ) -> None:
        object_ = (obj, action, iobj)
        # A slight hack: We remove trailing ellipsis and whitespace
        name = " → ".join(
            str(o).strip(".… ") for o in object_ if o is not None
        )
        objects.RunnableLeaf.__init__(self, object_, name)

    def __getstate__(self) -> dict[str, ty.Any]:
        state = dict(vars(self))
        state["object"] = [puid.get_unique_id(o) for o in self.object]
        return state

    def __setstate__(self, state: dict[str, ty.Any]) -> None:
        vars(self).update(state)
        objid, actid, iobjid = state["object"]
        obj = puid.resolve_unique_id(objid)
        assert isinstance(obj, Leaf)
        act = puid.resolve_action_id(actid, obj)
        iobj = puid.resolve_unique_id(iobjid)
        if (not obj or not act) or (iobj is None) != (iobjid is None):
            raise exceptions.InvalidDataError(f"Parts of {self} not restored")

        self.object[::] = [obj, act, iobj]

    def get_actions(self) -> ty.Iterator[Action]:
        yield actions.Perform()
        yield TimedPerform()

    def repr_key(self) -> ty.Any:
        return self

    def wants_context(self) -> bool:
        return True

    def run(self, ctx: commandexec.ExecutionToken | None = None) -> None:
        obj, action, iobj = self.object
        assert ctx and hasattr(ctx, "delegated_run")
        ctx.delegated_run(obj, action, iobj)

    def get_gicon(self) -> icons.GIcon | None:
        obj, action, _iobj = self.object
        return icons.ComposedIcon(obj.get_icon(), action.get_icon())


class _MultipleLeafContentSource(Source):
    def __init__(self, leaf: Leaf) -> None:
        Source.__init__(self, str(leaf))
        self.leaf = leaf

    def get_items(self) -> ty.Any:
        return self.leaf.object


class MultipleLeaf(Leaf):
    """A Leaf for the direct representation of many leaves. It is not
    a container or "source", it *is* the many leaves itself.

    The represented object is a sequence of Leaves."""

    serializable = 1

    def __init__(self, obj: ty.Any, name: str = _("Multiple Objects")) -> None:
        # modifying the list of objects is strictly forbidden
        robj = list(itertools.unique_iterator(obj))
        Leaf.__init__(self, robj, name)

    def get_multiple_leaf_representation(self) -> ty.Sequence[Leaf]:
        return ty.cast("ty.Sequence[Leaf]", self.object)

    def __getstate__(self) -> dict[str, ty.Any]:
        state = dict(vars(self))
        state["object"] = [puid.get_unique_id(o) for o in self.object]
        return state

    def __setstate__(self, state: dict[str, ty.Any]) -> None:
        vars(self).update(state)
        objs = []
        for id_ in state["object"]:
            if (obj := puid.resolve_unique_id(id_)) is not None:
                objs.append(obj)
            else:
                raise exceptions.InvalidDataError(
                    f"{id_} could not be restored!"
                )

        self.object[::] = objs

    def has_content(self) -> bool:
        return True

    def content_source(self, alternate: bool = False) -> Source:
        return _MultipleLeafContentSource(self)

    def get_description(self) -> str:
        num = len(self.object)
        return ngettext("%s object", "%s objects", num) % (num,)

    def get_icon_name(self) -> str:
        return "kupfer-object-multiple"



# pylint:disable=protected-access
# type:ignore

"""
Tests for validators.
"""

import unittest

from kupfer.support import validators as v


class TestValidateNetloc(unittest.TestCase):
    def test_ipv4(self):
        self.assertTrue(v.validate_netloc("127.0.0.1"))
        self.assertTrue(v.validate_netloc("0.0.0.0"))
        self.assertTrue(v.validate_netloc("255.255.255.255"))

        self.assertTrue(v.validate_netloc("@127.0.0.1"))
        self.assertTrue(v.validate_netloc("username@127.0.0.1"))
        self.assertTrue(v.validate_netloc("username:pass@127.0.0.1"))

        self.assertTrue(v.validate_netloc("127.0.0.1:123"))
        self.assertTrue(v.validate_netloc("@127.0.0.1:2314"))
        self.assertTrue(v.validate_netloc("username@127.0.0.1:21"))
        self.assertTrue(v.validate_netloc("username:pass@127.0.0.1:3"))

    def test_ipv4_neg(self):
        self.assertFalse(v.validate_netloc("127.0."))
        self.assertFalse(v.validate_netloc("127"))
        self.assertFalse(v.validate_netloc("276.0.0.0"))
        self.assertFalse(v.validate_netloc("255.255.255.255.244"))
        self.assertFalse(v.validate_netloc(":pass@127.0.0.1"))
        self.assertFalse(v.validate_netloc("@127.0.0.1:66666"))

    def test_ipv6(self):
        self.assertTrue(v.validate_netloc("[2001:db8:1::ab9:C:102]"))
        self.assertTrue(v.validate_netloc("[2001:db8:]"))
        self.assertTrue(v.validate_netloc("[::1]"))
        self.assertTrue(v.validate_netloc("[1:db8::1234:5678]"))
        self.assertTrue(v.validate_netloc("[::]"))

        self.assertTrue(v.validate_netloc("@[2001:db8:1::ab9:C:102]:12"))
        self.assertTrue(v.validate_netloc("user@[2001:db8:]:12"))
        self.assertTrue(v.validate_netloc("user:pass@[::1]:2131"))
        self.assertTrue(v.validate_netloc("[1:db8::1234:5678]:21312"))
        self.assertTrue(v.validate_netloc("[::]:1"))

    def test_ipv6_neg(self):
        # url need ipv6 address in [ ]
        self.assertFalse(v.validate_netloc("2001:db8:1::ab9:C:102"))

        self.assertFalse(v.validate_netloc("[2001]"))
        self.assertFalse(v.validate_netloc("[2:2:3:4:4:5:5:5:5:6]"))
        self.assertFalse(
            v.validate_netloc("[aaaa:bbbb:aaaa:bbbb:aaaa:bbbb:aaaa:bbbb:]")
        )
        self.assertFalse(
            v.validate_netloc("[aaaa:bbbb:aaaa:bbbb:aaaa:bbbb::aaaa:bbbb]")
        )
        self.assertFalse(
            v.validate_netloc("[aaaa:zbbb:aaaa:bbbb:aaaa:bbbb::azaa:bbbb]")
        )

    def test_hostname(self):
        self.assertTrue(v.validate_netloc("localhost"))
        self.assertTrue(v.validate_netloc("localhost:123"))
        self.assertTrue(v.validate_netloc("user@localhost"))
        self.assertTrue(v.validate_netloc("user:pass@localhost:123"))

        self.assertTrue(v.validate_netloc("www.abc.com"))
        self.assertTrue(v.validate_netloc("www.abc.com."))
        self.assertTrue(v.validate_netloc("www.abc-bc.com"))
        self.assertTrue(v.validate_netloc("www.ąśłóę.com"))
        self.assertTrue(v.validate_netloc("www.abc.com.pl"))
        self.assertTrue(v.validate_netloc("www.abc.com.pl:1234"))
        self.assertTrue(v.validate_netloc("user@www.abc.com.pl:1234"))
        self.assertTrue(v.validate_netloc("user:pass@www.abc.com.pl:1234"))

    def test_hostname_neg(self):
        self.assertTrue(v.validate_netloc("www.ab-c.com"))
        self.assertTrue(v.validate_netloc("www.abc--abc.com"))
        self.assertFalse(v.validate_netloc("anc*test.31231.com"))
        self.assertFalse(v.validate_netloc(".31231.com"))
        self.assertFalse(
            v.validate_netloc(
                ".123456789012345678901234567890123456789012345678901234567890.com"
            )
        )
        self.assertFalse(v.validate_netloc(".31231-.com"))
        self.assertFalse(v.validate_netloc("test@"))


class TestIsUrl(unittest.TestCase):
    def test_is_http_domain(self):
        self.assertTrue(v._is_http_domain("www.com"))
        self.assertTrue(v._is_http_domain("www.abc.com"))
        self.assertTrue(v._is_http_domain("www.abc.io"))
        self.assertTrue(v._is_http_domain("test.pl"))
        self.assertTrue(v._is_http_domain("test.com"))
        self.assertTrue(v._is_http_domain("localhost"))
        self.assertTrue(v._is_http_domain("ftp.localhost"))
        self.assertTrue(v._is_http_domain("abc.local"))
        self.assertTrue(v._is_http_domain("abc.local"))
        self.assertTrue(v._is_http_domain("abc.home.arpa"))

        self.assertFalse(v._is_http_domain("abcd"))
        self.assertFalse(v._is_http_domain("abc.xxxxxxxx"))
        self.assertFalse(v._is_http_domain("abc..local"))
        self.assertFalse(v._is_http_domain("abc.local."))
        self.assertFalse(v._is_http_domain(".abc.local"))

    def test_valid_http(self):
        self.assertEqual("https://www.abc.com", v.is_url("www.abc.com"))
        self.assertEqual("http://www.abc.com", v.is_url("http://www.abc.com"))
        self.assertEqual("http://abc.com", v.is_url("http://abc.com"))
        self.assertEqual(
            "https://abc.com/abc?test", v.is_url("abc.com/abc?test")
        )

        self.assertEqual("http://localhost", v.is_url("localhost"))
        self.assertEqual(
            "https://localhost/test", v.is_url("https://localhost/test")
        )

        self.assertEqual(
            "https://john.doe@www.example.com:123/forum/questions/"
            "?tag=networking&order=newest#top",
            v.is_url(
                "https://john.doe@www.example.com:123/forum/questions/"
                "?tag=networking&order=newest#top"
            ),
        )

        self.assertEqual(
            "https://example.com/tests/file%20name.pdf",
            v.is_url("https://example.com/tests/file%20name.pdf"),
        )

    def test_valid_mail(self):
        self.assertEqual(
            "mailto:user@example.com",
            v.is_url("mailto:user@example.com"),
        )
        # "mail:" is invalid schema for email, but still valid url...
        # special case: fix to mailto
        self.assertEqual(
            "mailto:user@example.com",
            v.is_url("mail:user@example.com"),
        )

    def test_valid_other(self):
        self.assertEqual(
            "ftp://abc:123@localhost/test",
            v.is_url("ftp://abc:123@localhost/test"),
        )
        self.assertEqual(
            "ftp://abc:123@ftp.localhost/test",
            v.is_url("abc:123@ftp.localhost/test"),
        )

        self.assertEqual(
            "sftp://sftp.localhost/",
            v.is_url("sftp://sftp.localhost/"),
        )
        self.assertEqual(
            "sftp://test@localhost",
            v.is_url("sftp://test@localhost"),
        )
        self.assertEqual(
            "sftp://test@sftp.localhost/",
            v.is_url("sftp://test@sftp.localhost/"),
        )

        self.assertEqual(
            "ldap://[2001:db8::7]/c=GB?objectClass?one",
            v.is_url(
                "ldap://[2001:db8::7]/c=GB?objectClass?one",
            ),
        )
        self.assertEqual(
            "news:comp.infosystems.www.servers.unix",
            v.is_url("news:comp.infosystems.www.servers.unix"),
        )
        self.assertEqual(
            "telnet://192.0.2.16:80/",
            v.is_url(
                "telnet://192.0.2.16:80/",
            ),
        )

    def test_valid_http_neg(self):
        self.assertIsNone(v.is_url("www.abc"))
        self.assertIsNone(v.is_url("abcdds"))
        self.assertIsNone(v.is_url("http daldkal alkl"))
        self.assertIsNone(v.is_url("com."))

    def test_short_netloc(self):
        self.assertEqual("sftp://localhost", v.is_url("sftp://localhost"))
        self.assertEqual("sftp://test", v.is_url("sftp://test"))

    def test_invalid(self):
        self.assertIsNone(v.is_url("http://kla alsdalk"))
        self.assertIsNone(v.is_url("http://www.^abc"))
        self.assertIsNone(v.is_url("http://www.abc|dsl"))


class TestIsEmail(unittest.TestCase):
    def test_is_valid_email(self):
        self.assertTrue(v.is_valid_email("test@ldldl.com"))
        self.assertTrue(v.is_valid_email("test1w123@ldldl.com.pl"))

        self.assertFalse(v.is_valid_email("test@"))
        self.assertFalse(v.is_valid_email("@ldldl.com.pl"))
        self.assertFalse(v.is_valid_email("ldldl.com.pl"))



"""
Module for configuration and misc things
"""

from __future__ import annotations

import os
import typing as ty
from pathlib import Path

from xdg import BaseDirectory

try:
    from kupfer import version_subst  # type:ignore
except ImportError:
    version_subst = None

PACKAGE_NAME = "kupfer"

__all__ = (
    "ResourceLookupError",
    "get_cache_file",
    "get_cache_home",
    "get_config_file",
    "get_config_files",
    "get_data_dirs",
    "get_data_file",
    "get_data_home",
    "get_kupfer_env",
    "has_capability",
    "save_config_file",
    "save_data_file",
)


class ResourceLookupError(Exception):
    pass


def has_capability(cap: str) -> bool:
    """Check is @cap capability is not disabled by environment variable"""
    return not bool(os.getenv(f"KUPFER_NO_{cap}"))


def get_kupfer_env(name: str, default: str = "") -> str:
    """Get valaue of KUPFER_<name> environment variable or default"""
    return os.getenv(f"KUPFER_{name}", default)


def get_cache_home() -> str | None:
    """Directory where cache files should be put.  Guaranteed to exist."""
    cache_home = BaseDirectory.xdg_cache_home or os.path.expanduser("~/.cache")
    cache_dir = Path(cache_home, PACKAGE_NAME)
    if not cache_dir.exists():
        try:
            cache_dir.mkdir(mode=0o700)
        except OSError as exc:
            print(exc)
            return None

    return str(cache_dir)


def get_cache_file(path: tuple[str, ...] = ()) -> str | None:
    """Get file by @path from cache directory.  Return None when no exists."""
    cache_home = BaseDirectory.xdg_cache_home or os.path.expanduser("~/.cache")
    cache_path = Path(cache_home, *path)
    if not cache_path.exists():
        return None

    return str(cache_path)


def get_data_file(filename: str, package: str = PACKAGE_NAME) -> str:
    """Return path to @filename if it exists anywhere in the data paths, else
    raise ResourceLookupError."""
    if version_subst:
        first_datadir = os.path.join(version_subst.DATADIR, package)
    else:
        first_datadir = "./data"

    file_path = Path(first_datadir, filename)
    if file_path.exists():
        return str(file_path)

    for data_path in BaseDirectory.load_data_paths(package):
        file_path = Path(data_path, filename)
        if file_path.exists():
            return str(file_path)

    if package == PACKAGE_NAME:
        raise ResourceLookupError(f"Resource {filename} not found")

    raise ResourceLookupError(
        f"Resource {filename} in package {package} not found"
    )


def save_data_file(filename: str) -> str | None:
    """Return filename in the XDG data home directory, where the directory is
    guaranteed to exist."""
    if direc := BaseDirectory.save_data_path(PACKAGE_NAME):
        return os.path.join(direc, filename)

    return None


def get_data_home() -> str:
    """Directory where data is to be saved. Guaranteed to exist."""
    return BaseDirectory.save_data_path(PACKAGE_NAME)  # type: ignore


def get_data_dirs(
    name: str = "", package: str = PACKAGE_NAME
) -> ty.Iterable[str]:
    """Iterate over all data dirs of @name that exist."""
    return ty.cast(
        "ty.Iterable[str]",
        BaseDirectory.load_data_paths(os.path.join(package, name)),
    )


def get_config_file(filename: str, package: str = PACKAGE_NAME) -> str | None:
    """Return path to @package/@filename if it exists anywhere in the config
    paths, else return None"""
    return ty.cast(
        "ty.Union[str, None]",
        BaseDirectory.load_first_config(package, filename),
    )


def get_config_files(filename: str) -> ty.Iterable[str]:
    """Iterator to @filename in all config paths, with most important (takes
    precedence) files first."""
    return BaseDirectory.load_config_paths(PACKAGE_NAME, filename) or ()


def save_config_file(filename: str) -> str | None:
    """Return filename in the XDG data home directory, where the directory
    is guaranteed to exist."""
    if direc := BaseDirectory.save_config_path(PACKAGE_NAME):
        return os.path.join(direc, filename)

    return None



"""
Helper functions - common code for core module.

This file is a part of the program kupfer, which is
released under GNU General Public License v3 (or any later version),
see the main program file, and COPYING for details.
"""

from __future__ import annotations

import typing as ty

if ty.TYPE_CHECKING:
    from kupfer.obj import Leaf

__all__ = ("get_leaf_members", "is_multiple_leaf")


def get_leaf_members(leaf: Leaf) -> ty.Sequence[Leaf]:
    """Return an iterator to members of @leaf, if it is a multiple leaf."""

    if hasattr(leaf, "get_multiple_leaf_representation"):
        return leaf.get_multiple_leaf_representation()  # type: ignore

    return (leaf,)


def is_multiple_leaf(leaf: Leaf | None) -> bool:
    """Check is leaf represent multiple leaves."""
    return hasattr(leaf, "get_multiple_leaf_representation")



__kupfer_name__ = _("Volumes and Disks")
__kupfer_sources__ = ("VolumesSource",)
__description__ = _("Mounted volumes and disks")
__version__ = "2023-05-01"
__author__ = "US, KB"

import typing as ty

from gi.repository import Gio, GLib

from kupfer import launch
from kupfer.core import commandexec
from kupfer.obj import Action, FileLeaf, Leaf, OpenTerminal, Source
from kupfer.obj.fileactions import Open
from kupfer.ui import uiutils

if ty.TYPE_CHECKING:
    from gettext import gettext as _


_VOLUME_ICON_NAME = "drive-removable-media"


class Volume(FileLeaf):
    """
    The Volume class actually represents one instance
    of GIO's GMount (as long as it is mounted)
    """

    # NOTE: marking as non-serializable
    serializable = None

    def __init__(self, volume):
        self.volume = volume
        fil = self.volume.get_root()
        path = fil.get_path()
        super().__init__(obj=path, name=volume.get_name())
        self.kupfer_add_alias(fil.get_basename())

    def get_actions(self):
        yield Open()
        yield OpenTerminal()

        if self.volume.can_eject():
            yield Eject()

        if self.volume.can_unmount():
            yield Unmount()

    def is_valid(self):
        vmgr = Gio.VolumeMonitor.get()
        return self.volume in vmgr.get_mounts()

    def get_description(self):
        # TRANS: %s is path where volume is mounted
        return _(
            "Volume mounted at %s"
        ) % launch.get_display_path_for_bytestring(self.object)

    def get_gicon(self):
        return self.volume.get_icon()

    def get_icon_name(self):
        return _VOLUME_ICON_NAME


class VolumeNotMounted(Leaf):
    """The Volume class actually represents one instance of GIO's GMount that
    is not mounted anywhere."""

    # NOTE: marking as non-serializable
    serializable = None

    def __init__(self, volume, device):
        super().__init__(volume, name=volume.get_name())
        self.device = device

    def get_actions(self):
        if self.object.can_mount():
            yield Mount()

    def get_description(self):
        # TRANS: %s is name of device with volume
        return _("Volume on %s") % self.device.get_name()

    def get_gicon(self):
        return self.device.get_icon()

    def get_icon_name(self):
        return _VOLUME_ICON_NAME


class Mount(Action):
    def __init__(self, name=None):
        super().__init__(name or _("Mount"))

    def mount_callback(
        self,
        volume: Gio.Volume,
        async_result: ty.Any,
        ctx: commandexec.ExecutionToken,
    ) -> None:
        try:
            volume.mount_finish(async_result)
        except GLib.Error:
            ctx.register_late_error()
        else:
            uiutils.show_notification(
                _("Mount finished"),
                # TRANS: %s is name of volume
                _('"%s" was successfully mounted') % volume.get_name(),
                icon_name=_VOLUME_ICON_NAME,
            )
            ctx.register_late_result(commandexec.ActionResultRefresh)

    def wants_context(self):
        return True

    def activate(self, leaf, iobj=None, ctx=None):
        assert ctx
        vol = leaf.object
        if vol.can_mount():
            vol.mount(
                Gio.MountMountFlags.NONE,
                None,
                None,
                self.mount_callback,
                ctx,
            )

    def get_description(self):
        return _("Mount this volume")


class Unmount(Action):
    # prefer eject over umount
    rank_adjust = -5

    def __init__(self, name=None):
        super().__init__(name or _("Unmount"))

    def unmount_callback(
        self,
        mount: Gio.Mount,
        async_result: ty.Any,
        ctx: commandexec.ExecutionToken,
    ) -> None:
        try:
            mount.unmount_with_operation_finish(async_result)
        except GLib.Error:
            # FIXME: argument
            ctx.register_late_error()
        else:
            self.success(mount.get_name())
            ctx.register_late_result(commandexec.ActionResultRefresh)

    def success(self, name: str) -> None:
        uiutils.show_notification(
            _("Unmount finished"),
            # TRANS: %s is name of volume
            _('"%s" was successfully unmounted') % name,
            icon_name=_VOLUME_ICON_NAME,
        )

    def wants_context(self):
        return True

    def activate(self, leaf, iobj=None, ctx=None):
        assert ctx

        if not leaf.is_valid():
            return

        vol = leaf.volume
        if vol.can_unmount():
            vol.unmount_with_operation(
                Gio.MountUnmountFlags.NONE,
                None,
                None,
                self.unmount_callback,
                ctx,
            )

    def get_description(self):
        return _("Unmount this volume")

    def get_icon_name(self):
        return "media-eject"


class Eject(Unmount):
    def __init__(self):
        super().__init__(_("Eject"))

    def get_description(self):
        return _("Unmount and eject this media")

    def activate(self, leaf, iobj=None, ctx=None):
        assert ctx

        if not leaf.is_valid():
            return

        vol = leaf.volume
        if vol.can_eject():
            vol.eject_with_operation(
                Gio.MountUnmountFlags.NONE,
                None,
                None,
                self.eject_callback,
                ctx,
            )

    def eject_callback(
        self,
        mount: Gio.Mount,
        async_result: ty.Any,
        ctx: commandexec.ExecutionToken,
    ) -> None:
        try:
            mount.eject_with_operation_finish(async_result)
        except GLib.Error:
            ctx.register_late_error()
        else:
            self.success(mount.get_name())
            ctx.register_late_result(commandexec.ActionResultRefresh)


class VolumesSource(Source):
    source_use_cache = False

    def __init__(self, name=_("Volumes and Disks")):
        self.volmon = None
        super().__init__(name)

    def initialize(self):
        self.volmon = Gio.VolumeMonitor.get()
        self.volmon.connect("mount-added", self._update)
        self.volmon.connect("mount-changed", self._update)
        self.volmon.connect("mount-removed", self._update)
        self.volmon.connect("drive-connected", self._update)
        self.volmon.connect("drive-changed", self._update)
        self.volmon.connect("drive-disconnected", self._update)

    def _update(self, *args):
        self.mark_for_update()
        GLib.timeout_add_seconds(1, self.mark_for_update)

    def finalize(self):
        del self.volmon

    def get_items(self):
        assert self.volmon
        # get_mounts gets all mounted removable media
        volumes = self.volmon.get_mounts()
        yield from map(Volume, volumes)

        # add unmounted removable devices
        for dev in self.volmon.get_connected_drives():
            if not dev.is_removable():
                continue

            for vol in dev.get_volumes():
                if vol.can_mount():
                    yield VolumeNotMounted(vol, dev)

    def get_description(self):
        return _("Mounted volumes and disks")

    def get_icon_name(self):
        return "drive-removable-media"

    def provides(self):
        yield Volume
        yield VolumeNotMounted



from __future__ import annotations

__kupfer_name__ = _("Firefox Keywords")
__kupfer_sources__ = ("KeywordsSource",)
__kupfer_text_sources__ = ("KeywordSearchSource",)
__kupfer_actions__ = ("SearchWithEngine",)
__description__ = _("Search the web with Firefox keywords")
__version__ = "2020.1"
__author__ = ""

import typing as ty
from urllib.parse import quote, urlparse

from kupfer import launch, plugin_support
from kupfer.obj import (
    Action,
    Leaf,
    OpenUrl,
    RunnableLeaf,
    Source,
    TextLeaf,
    TextSource,
)
from kupfer.obj.apps import AppLeafContentMixin
from kupfer.obj.helplib import FilesystemWatchMixin
from kupfer.plugin._firefox_support import (
    get_firefox_home_file,
    query_database,
)

if ty.TYPE_CHECKING:
    from gettext import gettext as _


__kupfer_settings__ = plugin_support.PluginSettings(
    {
        "key": "default",
        "label": _("Default for ?"),
        "type": str,
        "value": "https://www.google.com/search?ie=UTF-8&q=%s",
    },
    {
        "key": "profile",
        "label": _("Firefox profile name or path"),
        "type": str,
        "value": "",
        "helper": "choose_directory",
    },
)


def _url_domain(text: str) -> str:
    components = list(urlparse(text))
    return "".join(components[1:2])


class Keyword(Leaf):
    def __init__(self, title: str, kw: str, url: str) -> None:
        title = title or _url_domain(url)
        name = f"{kw} ({title})"
        super().__init__(url, name)
        self.keyword = kw
        self.is_search = "%s" in url

    def get_actions(self):
        if self.is_search:
            yield SearchFor()
        else:
            yield OpenUrl()

    def get_description(self):
        return self.object

    def get_icon_name(self):
        return "text-html"

    def get_text_representation(self):
        return self.object


_KEYWORDS_SQL = """
SELECT distinct moz_places.url, moz_places.title, moz_keywords.keyword
FROM moz_places, moz_keywords
WHERE moz_places.id = moz_keywords.place_id
"""


class KeywordsSource(AppLeafContentMixin, Source, FilesystemWatchMixin):
    appleaf_content_id = ("firefox", "firefox-esr")
    source_scan_interval: int = 3600

    instance: KeywordsSource | None = None

    def __init__(self):
        self.monitor_token = None
        super().__init__(_("Firefox Keywords"))

    def initialize(self):
        KeywordsSource.instance = self
        profile = __kupfer_settings__["profile"]
        if ff_home := get_firefox_home_file("", profile):
            self.monitor_token = self.monitor_directories(str(ff_home))

    def finalize(self):
        KeywordsSource.instance = None

    def monitor_include_file(self, gfile):
        return gfile and gfile.get_basename() == "lock"

    def mark_for_update(self, postpone=False):
        super().mark_for_update(postpone=True)

    def get_items(self):
        """Query the firefox places bookmark database"""
        fpath = get_firefox_home_file(
            "places.sqlite", __kupfer_settings__["profile"]
        )
        if not fpath:
            return []

        return [
            Keyword(title or url, kw, url)
            for url, title, kw in query_database(str(fpath), _KEYWORDS_SQL)
        ]

    def get_description(self):
        return None

    def get_icon_name(self):
        return "web-browser"

    def provides(self):
        yield Keyword


class SearchWithEngine(Action):
    """TextLeaf -> SearchWithEngine -> Keyword"""

    action_accelerator = "s"

    def __init__(self):
        Action.__init__(self, _("Search With..."))

    def activate(self, leaf, iobj=None, ctx=None):
        assert iobj
        url = iobj.object
        _do_search_engine(leaf.object, url)

    def item_types(self):
        yield TextLeaf

    def requires_object(self):
        return True

    def object_types(self):
        yield Keyword

    def valid_object(self, obj, for_item):
        return obj.is_search

    def object_source(self, for_item=None):
        return KeywordsSource()

    def get_description(self):
        return _("Search the web with Firefox keywords")

    def get_icon_name(self):
        return "edit-find"


class SearchFor(Action):
    """Keyword -> SearchFor -> TextLeaf

    This is the opposite action to SearchWithEngine
    """

    action_accelerator = "s"

    def __init__(self):
        Action.__init__(self, _("Search For..."))

    def activate(self, leaf, iobj=None, ctx=None):
        assert iobj
        url = leaf.object
        terms = iobj.object
        _do_search_engine(terms, url)

    def item_types(self):
        yield Keyword

    def requires_object(self):
        return True

    def object_types(self):
        yield TextLeaf

    def object_source(self, for_item=None):
        return TextSource(placeholder=_("Search Terms"))

    def valid_object(self, obj, for_item):
        # NOTE: Using exact class to skip subclasses
        return type(obj) == TextLeaf  # pylint: disable=unidiomatic-typecheck # noqa: E721

    def get_description(self):
        return _("Search the web with Firefox keywords")

    def get_icon_name(self):
        return "edit-find"


class KeywordSearchSource(TextSource):
    """This source allow search using firefox keywords.

    To use - i.e. define bookmarks for "https://hn.algolia.com/?q=%s" with keyword
    "hn" and then you can search by '?hn something to search'

    """

    def __init__(self):
        super().__init__(_("Firefox Keywords (?-source)"))

    def get_text_items(self, text):
        if not text.startswith("?"):
            return

        parts = text[1:].split(maxsplit=1)
        if len(parts) < 1:
            return

        query = parts[1] if len(parts) > 1 else ""
        assert KeywordsSource.instance
        for keyword in KeywordsSource.instance.get_leaves():
            assert isinstance(keyword, Keyword)
            if keyword.is_search and keyword.keyword == parts[0]:
                yield SearchWithKeyword(keyword, query)
                return

        if default := __kupfer_settings__["default"].strip():
            if "%s" not in default:
                default += "%s"

            yield SearchWithKeyword(Keyword("", "", default), text[1:])

    def get_description(self):
        return None

    def get_icon_name(self):
        return "web-browser"

    def provides(self):
        yield SearchWithKeyword

    def get_rank(self):
        return 80


class SearchWithKeyword(RunnableLeaf):
    object: tuple[Keyword, str]

    def __init__(self, keyword: Keyword, text: str):
        super().__init__((keyword, text), _('Search for "%s"') % (text,))

    def run(self, ctx=None):
        kwl = self.keyword_leaf
        _do_search_engine(self.query, kwl.object)

    @property
    def keyword_leaf(self) -> Keyword:
        return self.object[0]

    @property
    def query(self) -> str:
        return self.object[1]

    def get_icon_name(self):
        return "web-browser"

    def get_description(self):
        return _("Search using %s") % self.keyword_leaf

    def get_text_representation(self):
        kwl = self.keyword_leaf
        return _query_url(self.query, kwl.object)


def _do_search_engine(
    terms: str, search_url: str, encoding: str = "UTF-8"
) -> None:
    """Show an url searching for @search_url with @terms"""
    launch.show_url(_query_url(terms, search_url))


def _query_url(terms: str, search_url: str) -> str:
    """Show an url searching for @search_url with @terms"""
    return search_url.replace("%s", quote(terms))



"""This plugin opens WhatsApp Web in the browser, and can send messages to
numbers that are not scheduled
"""

__kupfer_name__ = _("WhatsApp Web")
__kupfer_sources__ = ()
__kupfer_actions__ = ("WhatsApp",)
__description__ = _(
    """Open a new number in WhatsApp Web

The number format must respect:
"Country Code" + "Area Code" + "Number"

Example for a New York number: 12129999999
Country Code: 1
Area Code: 212
Number: 9999999

WhatsApp Web will open in the browser.

For help visit https://faq.whatsapp.com/general/chats/how-to-use-click-to-chat/

"""
)
__version__ = "1.0"
__author__ = "Leonardo Masuero <leom255255@gmail.com>"

import typing as ty
from contextlib import suppress

from kupfer import launch
from kupfer.obj import Action, TextLeaf

if ty.TYPE_CHECKING:
    from gettext import gettext as _


class WhatsApp(Action):
    def __init__(self):
        Action.__init__(self, _("WhatsApp Web"))

    def activate(self, leaf, iobj=None, ctx=None):
        url_w = "https://web.whatsapp.com/send?phone="
        url_number = url_w + leaf.object
        launch.show_url(url_number)

    def item_types(self):
        yield TextLeaf

    def valid_for_item(self, leaf):
        with suppress(BaseException):
            text = leaf.object
            return bool(text) and int(text)

    def get_description(self):
        return _("Send a WhatsApp to a new number.")

    def get_icon_name(self):
        return "message-new"



# pylint: disable=invalid-name
# ruff: noqa: N802
"""
This module has a singleton Service for dbus callbacks,
and ensures there is only one unique service in the Session
"""

from __future__ import annotations

import typing as ty

from gi.repository import GObject

try:
    import dbus
    from dbus.mainloop.glib import DBusGMainLoop

    DBusGMainLoop(set_as_default=True)
    import dbus.service
    from dbus.gi_service import ExportedGObject

    dbus.mainloop.glib.threads_init()

    # if dbus unavailable print the exception here
    # but further actions (register) will fail without warning
    _SESSION_BUS = dbus.Bus()
except (ImportError, dbus.exceptions.DBusException) as exc:
    _SESSION_BUS = None
    print(exc)

from kupfer.ui import uievents


class AlreadyRunningError(Exception):
    """Service already available on the bus Exception"""


class NoConnectionError(Exception):
    """Not possible to establish connection for callbacks"""


_SERVER_NAME: ty.Final = "se.kaizer.kupfer"
_INTERFACE_NAME: ty.Final = "se.kaizer.kupfer.Listener"
_OBJECT_NAME: ty.Final = "/interface"

_SERVER_NAME_NEW: ty.Final = "io.github.kupferlauncher"
_INTERFACE_NAME_NEW: ty.Final = "io.github.kupferlauncher.Listener"
_OBJECT_NAME_NEW: ty.Final = "/io/github/kupferlauncher"


class Service(ExportedGObject):  # type:ignore
    def __init__(self):
        """Create a new Kupfer service on the Session Bus

        Raises NoConnectionError, AlreadyRunningError
        """
        if not _SESSION_BUS:
            raise NoConnectionError

        if _SESSION_BUS.name_has_owner(_SERVER_NAME):
            raise AlreadyRunningError

        bus_name = dbus.service.BusName(_SERVER_NAME, bus=_SESSION_BUS)
        super().__init__(
            conn=_SESSION_BUS, object_path=_OBJECT_NAME, bus_name=bus_name
        )

    def unregister(self):
        if _SESSION_BUS:
            _SESSION_BUS.release_name(_SERVER_NAME)

    @dbus.service.method(_INTERFACE_NAME)
    def Present(self):
        self.PresentOnDisplay("", "")

    @dbus.service.method(_INTERFACE_NAME, in_signature="ay", byte_arrays=True)
    def PresentWithStartup(self, notify_id):
        self.PresentOnDisplay("", notify_id)

    @dbus.service.method(
        _INTERFACE_NAME, in_signature="ayay", byte_arrays=True
    )
    def PresentOnDisplay(self, display, notify_id):
        with uievents.using_startup_notify_id(notify_id) as time:
            self.emit("present", display, time)

    @dbus.service.method(_INTERFACE_NAME)
    def ShowHide(self):
        self.emit("show-hide", "", 0)

    @dbus.service.method(_INTERFACE_NAME, in_signature="s")
    def PutText(self, text):
        self.PutTextOnDisplay(text, "", "")

    @dbus.service.method(
        _INTERFACE_NAME, in_signature="sayay", byte_arrays=True
    )
    def PutTextOnDisplay(self, text, display, notify_id):
        with uievents.using_startup_notify_id(notify_id) as time:
            self.emit("put-text", text, display, time)

    @dbus.service.method(_INTERFACE_NAME, in_signature="as")
    def PutFiles(self, fileuris):
        self.PutFilesOnDisplay(fileuris, "", "")

    @dbus.service.method(
        _INTERFACE_NAME, in_signature="asayay", byte_arrays=True
    )
    def PutFilesOnDisplay(self, fileuris, display, notify_id):
        # files sent with dbus-send from kupfer have a custom comma
        # escape that we have to unescape here
        fileuris[:] = [f.replace("%%kupfercomma%%", ",") for f in fileuris]
        with uievents.using_startup_notify_id(notify_id) as time:
            self.emit("put-files", fileuris, display, time)

    @dbus.service.method(_INTERFACE_NAME, in_signature="s")
    def FindObject(self, qpfer):
        self.FindObjectOnDisplay(qpfer, "", "")

    @dbus.service.method(_INTERFACE_NAME, in_signature="sss")
    def FindObjectOnDisplay(self, qpfer, display, notify_id):
        with uievents.using_startup_notify_id(notify_id) as time:
            self.emit("find-object", qpfer, display, time)

    @dbus.service.method(_INTERFACE_NAME, in_signature="s")
    def ExecuteFile(self, filepath):
        self.ExecuteFileOnDisplay(filepath, "", "")

    @dbus.service.method(
        _INTERFACE_NAME, in_signature="sayay", byte_arrays=True
    )
    def ExecuteFileOnDisplay(self, filepath, display, notify_id):
        with uievents.using_startup_notify_id(notify_id) as time:
            self.emit("execute-file", filepath, display, time)

    @dbus.service.method(
        _INTERFACE_NAME, in_signature="sayay", byte_arrays=True
    )
    def RelayKeysFromDisplay(self, keystring, display, notify_id):
        with uievents.using_startup_notify_id(notify_id) as time:
            self.emit("relay-keys", keystring, display, time)

    @dbus.service.method(
        _INTERFACE_NAME,
        in_signature=None,
        out_signature="as",
        byte_arrays=True,
    )
    def GetBoundKeys(self):
        # pylint: disable=import-outside-toplevel
        from kupfer.ui import keybindings

        return keybindings.get_all_bound_keys()

    @dbus.service.signal(_INTERFACE_NAME, signature="sb")
    def BoundKeyChanged(self, keystr, is_bound):
        pass

    @dbus.service.method(_INTERFACE_NAME)
    def Quit(self):
        self.emit("quit")


# Signature: displayname, timestamp
GObject.signal_new(
    "present",
    Service,
    GObject.SignalFlags.RUN_LAST,
    GObject.TYPE_BOOLEAN,
    (GObject.TYPE_STRING, GObject.TYPE_UINT),
)

# Signature: displayname, timestamp
GObject.signal_new(
    "show-hide",
    Service,
    GObject.SignalFlags.RUN_LAST,
    GObject.TYPE_BOOLEAN,
    (GObject.TYPE_STRING, GObject.TYPE_UINT),
)

# Signature: text, displayname, timestamp
GObject.signal_new(
    "put-text",
    Service,
    GObject.SignalFlags.RUN_LAST,
    GObject.TYPE_BOOLEAN,
    (GObject.TYPE_STRING, GObject.TYPE_STRING, GObject.TYPE_UINT),
)

# Signature: filearray, displayname, timestamp
GObject.signal_new(
    "put-files",
    Service,
    GObject.SignalFlags.RUN_LAST,
    GObject.TYPE_BOOLEAN,
    (GObject.TYPE_PYOBJECT, GObject.TYPE_STRING, GObject.TYPE_UINT),
)

# Signature: string, displayname, timestamp
GObject.signal_new(
    "find-object",
    Service,
    GObject.SignalFlags.RUN_LAST,
    GObject.TYPE_BOOLEAN,
    (GObject.TYPE_STRING, GObject.TYPE_STRING, GObject.TYPE_UINT),
)

# Signature: fileuri, displayname, timestamp
GObject.signal_new(
    "execute-file",
    Service,
    GObject.SignalFlags.RUN_LAST,
    GObject.TYPE_BOOLEAN,
    (GObject.TYPE_STRING, GObject.TYPE_STRING, GObject.TYPE_UINT),
)

# Signature: ()
GObject.signal_new(
    "quit", Service, GObject.SignalFlags.RUN_LAST, GObject.TYPE_BOOLEAN, ()
)

# Signature: keystring, displayname, timestamp
GObject.signal_new(
    "relay-keys",
    Service,
    GObject.SignalFlags.RUN_LAST,
    GObject.TYPE_BOOLEAN,
    (GObject.TYPE_STRING, GObject.TYPE_STRING, GObject.TYPE_UINT),
)


class ServiceNew(ExportedGObject):  # type: ignore
    def __init__(self):
        """Create a new Kupfer service on the Session Bus

        Raises NoConnectionError, AlreadyRunningError
        """
        if not _SESSION_BUS:
            raise NoConnectionError

        if _SESSION_BUS.name_has_owner(_SERVER_NAME_NEW):
            raise AlreadyRunningError

        bus_name = dbus.service.BusName(_SERVER_NAME_NEW, bus=_SESSION_BUS)
        super().__init__(
            conn=_SESSION_BUS, object_path=_OBJECT_NAME_NEW, bus_name=bus_name
        )

    def unregister(self):
        if _SESSION_BUS:
            _SESSION_BUS.release_name(_SERVER_NAME)

    @dbus.service.method(_INTERFACE_NAME_NEW)
    def Present(self):
        self.PresentOnDisplay("", "")

    @dbus.service.method(_INTERFACE_NAME_NEW, in_signature="ss")
    def PresentOnDisplay(self, display, notify_id):
        with uievents.using_startup_notify_id(notify_id) as time:
            self.emit("present", display, time)

    @dbus.service.method(_INTERFACE_NAME_NEW)
    def ShowHide(self):
        self.ShowHideOnDisplay("", "")

    @dbus.service.method(_INTERFACE_NAME_NEW, in_signature="ss")
    def ShowHideOnDisplay(self, display, notify_id):
        with uievents.using_startup_notify_id(notify_id) as time:
            self.emit("show-hide", display, time)

    @dbus.service.method(_INTERFACE_NAME_NEW, in_signature="s")
    def PutText(self, text):
        self.PutTextOnDisplay(text, "", "")

    @dbus.service.method(_INTERFACE_NAME_NEW, in_signature="sss")
    def PutTextOnDisplay(self, text, display, notify_id):
        with uievents.using_startup_notify_id(notify_id) as time:
            self.emit("put-text", text, display, time)

    @dbus.service.method(_INTERFACE_NAME_NEW, in_signature="as")
    def PutFiles(self, fileuris):
        self.PutFilesOnDisplay(fileuris, "", "")

    @dbus.service.method(_INTERFACE_NAME_NEW, in_signature="asss")
    def PutFilesOnDisplay(self, fileuris, display, notify_id):
        # files sent with dbus-send from kupfer have a custom comma
        # escape that we have to unescape here
        fileuris[:] = [f.replace("%%kupfercomma%%", ",") for f in fileuris]
        with uievents.using_startup_notify_id(notify_id) as time:
            self.emit("put-files", fileuris, display, time)

    @dbus.service.method(_INTERFACE_NAME_NEW, in_signature="s")
    def FindObject(self, qpfer):
        self.FindObjectOnDisplay(qpfer, "", "")

    @dbus.service.method(_INTERFACE_NAME_NEW, in_signature="sss")
    def FindObjectOnDisplay(self, qpfer, display, notify_id):
        with uievents.using_startup_notify_id(notify_id) as time:
            self.emit("find-object", qpfer, display, time)

    @dbus.service.method(_INTERFACE_NAME_NEW, in_signature="s")
    def ExecuteFile(self, filepath):
        self.ExecuteFileOnDisplay(filepath, "", "")

    @dbus.service.method(_INTERFACE_NAME_NEW, in_signature="sss")
    def ExecuteFileOnDisplay(self, filepath, display, notify_id):
        with uievents.using_startup_notify_id(notify_id) as time:
            self.emit("execute-file", filepath, display, time)

    @dbus.service.method(_INTERFACE_NAME_NEW)
    def Quit(self):
        self.emit("quit")


# Signature: displayname, timestamp
GObject.signal_new(
    "present",
    ServiceNew,
    GObject.SignalFlags.RUN_LAST,
    GObject.TYPE_BOOLEAN,
    (GObject.TYPE_STRING, GObject.TYPE_UINT),
)

# Signature: displayname, timestamp
GObject.signal_new(
    "show-hide",
    ServiceNew,
    GObject.SignalFlags.RUN_LAST,
    GObject.TYPE_BOOLEAN,
    (GObject.TYPE_STRING, GObject.TYPE_UINT),
)

# Signature: text, displayname, timestamp
GObject.signal_new(
    "put-text",
    ServiceNew,
    GObject.SignalFlags.RUN_LAST,
    GObject.TYPE_BOOLEAN,
    (GObject.TYPE_STRING, GObject.TYPE_STRING, GObject.TYPE_UINT),
)

# Signature: filearray, displayname, timestamp
GObject.signal_new(
    "put-files",
    ServiceNew,
    GObject.SignalFlags.RUN_LAST,
    GObject.TYPE_BOOLEAN,
    (GObject.TYPE_PYOBJECT, GObject.TYPE_STRING, GObject.TYPE_UINT),
)

# Signature: string, displayname, timestamp
GObject.signal_new(
    "find-object",
    ServiceNew,
    GObject.SignalFlags.RUN_LAST,
    GObject.TYPE_BOOLEAN,
    (GObject.TYPE_STRING, GObject.TYPE_STRING, GObject.TYPE_UINT),
)

# Signature: fileuri, displayname, timestamp
GObject.signal_new(
    "execute-file",
    ServiceNew,
    GObject.SignalFlags.RUN_LAST,
    GObject.TYPE_BOOLEAN,
    (GObject.TYPE_STRING, GObject.TYPE_STRING, GObject.TYPE_UINT),
)

# Signature: ()
GObject.signal_new(
    "quit", ServiceNew, GObject.SignalFlags.RUN_LAST, GObject.TYPE_BOOLEAN, ()
)


# Note ServiceNew has the same signals, but doesn't actually implement them
# all (yet). Needs new design for wayland keyrelay.
# Signature: keystring, displayname, timestamp
GObject.signal_new(
    "relay-keys",
    ServiceNew,
    GObject.SignalFlags.RUN_LAST,
    GObject.TYPE_BOOLEAN,
    (GObject.TYPE_STRING, GObject.TYPE_STRING, GObject.TYPE_UINT),
)



# pylint: disable=protected-access
"""
This file is a part of the program kupfer, which is
released under GNU General Public License v3 (or any later version),
see the main program file, and COPYING for details.
"""

import unittest

from kupfer.support import desktop_parse


class TestUnescape(unittest.TestCase):
    def test1(self):
        self.assertEqual(
            desktop_parse._unescape(r'"This \\$ \\\\ \s\\\\"'),
            '"This \\$ \\\\  \\\\"',
        )
        self.assertEqual(desktop_parse._unescape(r"\t\s\\\\"), "\t \\\\")



from __future__ import annotations

import itertools
import operator
import typing as ty

from kupfer.core import learn, relevance
from kupfer.obj.base import Action, Leaf

__all__ = (
    "Rankable",
    "add_bonus_to_action",
    "add_bonus_to_objects",
    "add_rank_to_objects",
    "find_best_sort",
    "make_rankables",
    "score_actions",
    "score_objects",
    "wrap_rankable",
)

# RankableObject is type of object that can be put in Rankable.
RankableObject = ty.Union[Leaf, Action]


def make_rankables(
    itr: ty.Iterable[RankableObject], rank: int = 0
) -> ty.Iterable[Rankable]:
    """Create Rankable from some KupferObject:w"""
    return (Rankable(str(obj), obj, rank) for obj in itr)


def wrap_rankable(obj: Leaf, rank: int = 0) -> Rankable:
    return Rankable(str(obj), obj, rank)


class Rankable:
    """Rankable has an object (represented item), value (determines rank)
    and an associated rank."""

    # To save memory with (really) many Rankables
    __slots__ = ("aliases", "object", "rank", "value")

    def __init__(self, value: str, obj: RankableObject, rank: int = 0) -> None:
        self.rank: int = rank
        self.value: str = value
        self.object: RankableObject = obj
        self.aliases: ty.Collection[str] = getattr(obj, "name_aliases", ())

    def __str__(self):
        return f"{self.rank:.2f}: {self.value!r}, {self.object!r}"

    def __repr__(self):
        return f"<Rankable {self} repres {self.object!r} at {id(self):x}>"


def add_bonus_to_objects(
    rankables: ty.Iterable[Rankable], key: str, extra_bonus: int = 0
) -> ty.Iterator[Rankable]:
    """
    Increment rank of each item in `rankables` for mnemonic score for key and
    `extra_bonus`.
    """
    get_record_score = learn.get_record_score
    for obj in rankables:
        obj.rank += get_record_score(obj.object, key) + extra_bonus
        yield obj


def add_bonus_to_action(
    rankables: ty.Iterable[Rankable], key: str
) -> ty.Iterator[Rankable]:
    """
    generator of @rankables that have mnemonics for @key

    Add bonus for mnemonics and rank_adjust

    rank is added to prev rank, all items are yielded"""
    get_record_score = learn.get_record_score
    for obj in rankables:
        obj.rank += get_record_score(obj.object, key) + obj.object.rank_adjust
        yield obj


def add_rank_to_objects(
    rankables: ty.Iterable[Rankable], rank: int
) -> ty.Iterator[Rankable]:
    """
    Add @rank to rank of all @rankables.

    rankables: Iterable[Rankable] - updated
    rank: Fixed rank
    """
    for obj in rankables:
        obj.rank += rank
        yield obj


def score_objects(
    rankables: ty.Iterable[Rankable], key: str
) -> ty.Iterator[Rankable]:
    """
    rankables: List[Rankable]

    Prune rankables that score low for the key.
    """
    key = key.lower()
    _score = relevance.score_single if len(key) == 1 else relevance.score

    for rankable in rankables:
        # Rank object
        rank = int(_score(rankable.value, key) * 100)
        if rank < 90:  # noqa:PLR2004
            # consider aliases and change rb.value if alias is better
            # aliases rank lower so that value is chosen when close
            arank_value = max(
                ((_score(alias, key), alias) for alias in rankable.aliases),
                default=None,
            )
            if arank_value:
                arank, value = arank_value
                arank *= 95
                if arank > rank:
                    rankable.value = value
                    rank = int(arank)

        rankable.rank = rank

        if rankable.rank > 10:  # noqa:PLR2004
            yield rankable


def score_actions(
    rankables: ty.Iterable[Rankable], for_leaf: Leaf | None
) -> ty.Iterator[Rankable]:
    """Alternative (rigid) scoring mechanism for objects,
    putting much more weight in rank_adjust."""
    get_record_score = learn.get_record_score
    for obj in rankables:
        obj_object = ty.cast("Action", obj.object)
        rank_adj = obj_object.rank_adjust + learn.get_correlation_bonus(
            obj_object, for_leaf
        )

        if rank_adj > 0:
            obj.rank = 50 + rank_adj + get_record_score(obj_object) // 2
        elif rank_adj == 0:
            obj.rank = get_record_score(obj_object)
        else:
            obj.rank = -50 + rank_adj + get_record_score(obj_object)

        yield obj


_rank_key = operator.attrgetter("rank")


def find_best_sort(
    rankables: ty.Iterable[Rankable],
) -> ty.Iterable[Rankable]:
    """Yield rankables in best rank first order.
    A special kind of lazy sort: simply find the best ranked item and yield
    it first, then if needed continue by sorting the rest.

    Note: this will duplicate the best item."""
    r1, r2 = itertools.tee(rankables, 2)
    maxval = max(r1, default=None, key=_rank_key)
    if maxval is None:
        return

    yield maxval
    yield from sorted(r2, key=_rank_key, reverse=True)



"""
This module contains internal and / or experimental Kupfer features.

These are not meant to be useful to "normal" users of Kupfer -- if they are,
they can be tested here before they migrate to a fitting plugin.
"""

import io

from kupfer import puid

# NOTE: Core imports
from kupfer.core import learn, qfurl
from kupfer.obj import Action, Leaf, Source
from kupfer.obj.compose import ComposedLeaf
from kupfer.support import pretty
from kupfer.ui import uiutils

__kupfer_sources__ = ()
__kupfer_contents__ = ("ComposedSource",)
__kupfer_actions__ = ("DebugInfo", "Forget")
__description__ = __doc__
__author__ = "Ulrik Sverdrup <ulrik.sverdrup@gmail.com>"


class DebugInfo(Action):
    """Print debug info to terminal"""

    rank_adjust = -50

    def __init__(self):
        Action.__init__(self, "Debug Info")

    def activate(self, leaf, iobj=None, ctx=None):
        output = io.StringIO()

        def print_func(*args):
            print(" ".join(str(a) for a in args), file=output)
            pretty.print_debug("debug", *args)

        print_func("Debug info about", leaf)
        print_func(leaf, repr(leaf))

        def get_qfurl(leaf):
            try:
                return qfurl.Qfurl(leaf)
            except qfurl.QfurlError:
                pass
            return None

        def get_object_fields(leaf):
            return {
                "repr": leaf,
                "description": leaf.get_description(),
                "thumb": leaf.get_thumbnail(32, 32),
                "gicon": leaf.get_gicon(),
                "icon": leaf.get_icon(),
                "icon-name": leaf.get_icon_name(),
                "type": type(leaf),
                "module": leaf.__module__,
                "aliases": getattr(leaf, "name_aliases", None),
                "qfurl": get_qfurl(leaf),
                "puid": puid.get_unique_id(leaf),
            }

        def get_leaf_fields(leaf):
            base = get_object_fields(leaf)
            base.update(
                {
                    "object": leaf.object,
                    "object-type": type(leaf.object),
                    "content": leaf.content_source(),
                    "content-alt": leaf.content_source(alternate=True),
                    "builtin-actions": list(leaf.get_actions()),
                }
            )
            return base

        def get_source_fields(src):
            base = get_object_fields(src)
            base.update(
                {
                    "dynamic": src.is_dynamic(),
                    "sort": src.should_sort_lexically(),
                    "parent": src.get_parent(),
                    "leaf": src.get_leaf_repr(),
                    "provides": list(src.provides()),
                    "cached_items": type(src.cached_items),
                    "len": isinstance(src.cached_items, list)
                    and len(src.cached_items),
                }
            )
            return base

        def print_fields(fields):
            for field in sorted(fields):
                val = fields[field]
                rep = repr(val)
                print_func(f"{field:<15}:", rep)
                if str(val) not in rep:
                    print_func(f"{field:<15}:", val)

        leafinfo = get_leaf_fields(leaf)
        print_fields(leafinfo)
        if leafinfo["content"]:
            print_func("Content ============")
            print_fields(get_source_fields(leafinfo["content"]))

        if leafinfo["content"] != leafinfo["content-alt"]:
            print_func("Content-Alt ========")
            print_fields(get_source_fields(leafinfo["content-alt"]))

        print_func("Mnemonics ===========")
        print_func(learn._REGISTER.mnemonics.get(repr(leaf)))  # noqa: SLF001

        print_func("Actions correlation ===========")
        print_func(learn._REGISTER.correlations.get(repr(leaf)))  # noqa: SLF001
        print_func(f"Actions correlation for type {type(leaf)!r} ===========")
        print_func(learn._REGISTER.correlations.get(repr(type(leaf))))  # noqa: SLF001

        uiutils.show_text_result(output.getvalue())

    def get_description(self):
        return "Print debug output (for internal kupfer use)"

    def get_icon_name(self):
        return "emblem-system"

    def item_types(self):
        yield Leaf


class Forget(Action):
    rank_adjust = -10

    def __init__(self):
        Action.__init__(self, "Forget")

    def activate(self, leaf, iobj=None, ctx=None):
        # FIXME: This is a large, total, utter HACK
        if isinstance(leaf, ComposedLeaf):
            for obj in leaf.object:
                learn.unregister(repr(obj))

        if isinstance(leaf, ActionLeaf):
            learn.unregister(repr(leaf.object))
        else:
            learn.unregister(repr(leaf))

    def item_types(self):
        yield Leaf

    def get_description(self):
        return "Let Kupfer forget about this object"


class ActionLeaf(Leaf):
    def __init__(self, action):
        Leaf.__init__(self, action, str(action))

    def get_actions(self):
        act = self.object
        if not (hasattr(act, "requires_object") and act.requires_object()):
            yield Apply(act)

    def get_description(self):
        return self.object.get_description()

    def get_icon_name(self):
        return self.object.get_icon_name()


class Apply(Action):
    rank_adjust = 5

    def __init__(self, action):
        Action.__init__(self, "Apply To...")
        self.action = action

    def is_factory(self):
        return self.action.is_factory()

    def has_result(self):
        return self.action.has_result()

    def is_async(self):
        return self.action.is_async()

    def requires_object(self):
        return True

    def object_types(self):
        return self.action.item_types()

    def valid_object(self, obj, for_item=None):
        return self.action.valid_for_item(obj)

    def wants_context(self):
        return self.action.wants_context()

    def activate(self, leaf, iobj=None, ctx=None):
        return self.action.activate(iobj, ctx=ctx)


class ComposedSource(Source):
    """Decorated ComposedLeaf with a Source that shows the contents of
    Composed Commands"""

    def __init__(self, leaf):
        Source.__init__(self, "Composed Command")
        self.leaf = leaf

    def get_items(self):
        obj = self.leaf.object
        yield self.leaf.object[0]
        yield ActionLeaf(obj[1])
        if self.leaf.object[2] is not None:
            yield self.leaf.object[2]

    def repr_key(self):
        return self.leaf.repr_key()

    @classmethod
    def decorates_type(cls):
        return ComposedLeaf

    @classmethod
    def decorate_item(cls, leaf):
        return cls(leaf)



"""
Load directories from zoxide (https://github.com/ajeetdsouza/zoxide)
with configured, minimal score.

Optionally, some directories can be excluded (by path) and only existing
files may be presented (this may slowdown loading when turned on).
"""

__kupfer_name__ = _("Zoxide Directories")
__kupfer_sources__ = ("ZoxideDirSource",)
__description__ = _("Load top directories from zoxide database")
__version__ = "2023-04-02"
__author__ = "Karol Będkowski <karol.bedkowski@gmail.com>"

import os.path
import subprocess
import typing as ty

from kupfer import config, icons, plugin_support
from kupfer.core.datactrl import DataController
from kupfer.obj import Action, FileLeaf, Leaf, Source, fileactions
from kupfer.obj.helplib import FilesystemWatchMixin

if ty.TYPE_CHECKING:
    from gettext import gettext as _


__kupfer_settings__ = plugin_support.PluginSettings(
    {
        "key": "exclude",
        "label": _("Exclude directories:"),
        "type": list,
        "value": [],
        "helper": "choose_directory",
    },
    {
        "key": "min_score",
        "label": _("Minimal score:"),
        "type": int,
        "value": 1,
        "min": 1,
        "tooltip": _(
            "Load only directories that score is equal "
            "or higher than configured minimum"
        ),
    },
    {
        "key": "max_items",
        "label": _("Load limit:"),
        "type": int,
        "value": 50,
        "min": 1,
        "tooltip": _("Maximal number of directories to load"),
    },
    {
        "key": "existing",
        "label": _("Show only existing directories"),
        "type": bool,
        "value": True,
    },
    {
        "key": "record_enabled",
        "label": _("Update database after activate"),
        "type": bool,
        "value": False,
        "tooltip": _(
            "When enabled open files and directories will update zoxide database"
        ),
    },
)

plugin_support.check_command_available("zoxide")


class LaunchRecorder:
    """Launch recorder listen for 'launched-action' signals after Open  action on
    FileLeaves and update zoxide about access to file folder."""

    def __init__(self):
        self._enabled = False
        self._cb_pointer = None

    def connect(self):
        data_controller = DataController.instance()
        self._cb_pointer = data_controller.connect(
            "launched-action", self._on_launched_action
        )
        __kupfer_settings__.connect(
            "plugin-setting-changed", self._on_setting_changed
        )
        self._enabled = __kupfer_settings__["record_enabled"]

    def disconnect(self):
        if self._cb_pointer:
            data_controller = DataController.instance()
            data_controller.disconnect(self._cb_pointer)
            self._cb_pointer = None

    def _on_launched_action(
        self, sender: ty.Any, leaf: Leaf, action: Action, *_args: ty.Any
    ) -> None:
        if not self._enabled:
            return

        if not isinstance(leaf, FileLeaf):
            return

        if not isinstance(action, fileactions.Open):
            return

        path = leaf.object
        if not leaf.is_dir():
            path = os.path.dirname(path)

        # check is path is excluded
        if any(map(path.startswith, __kupfer_settings__["exclude"])):
            return

        subprocess.run(["zoxide", "add", path], check=False)

    def _on_setting_changed(self, settings, key, value):
        if key == "record_enabled":
            self._enabled = bool(value)


_RECORDER = LaunchRecorder()


def initialize_plugin(plugin_name: str) -> None:
    _RECORDER.connect()


def finalize_plugin(plugin_name: str) -> None:
    _RECORDER.disconnect()


def _get_dirs(
    exclude: list[str], min_score: int, existing: bool, max_items: int
) -> ty.Iterator[str]:
    """Load folders with score from zoxide."""
    cmd = ["zoxide", "query", "--list", "--score"]
    if not existing:
        cmd.append("--all")

    with subprocess.Popen(cmd, stdout=subprocess.PIPE) as proc:
        stdout, _stderr = proc.communicate()
        lines = stdout.splitlines()

    for rownum, line in enumerate(lines):
        if rownum > max_items:
            return

        line = line.strip()  # noqa: PLW2901
        score, _dummy, dirpath = line.partition(b" ")
        if not dirpath:
            continue

        if float(score) < min_score:
            return

        path = dirpath.decode()
        # zoxide query not support multiple --exclude; so filter it here
        if any(map(path.startswith, exclude)):
            continue

        yield path


class ZoxideDirSource(Source, FilesystemWatchMixin):
    source_scan_interval: int = 3600
    rank_adjust = 20

    def __init__(self):
        super().__init__(name=_("Zoxide Directories"))
        self.monitor = None

    def initialize(self):
        zoxide_home = config.get_data_dirs("", "zoxide")
        self.monitor = self.monitor_directories(*zoxide_home)
        __kupfer_settings__.connect(
            "plugin-setting-changed", self._on_setting_changed
        )

    def monitor_include_file(self, gfile):
        return gfile and gfile.get_basename() == "zo.db"

    def get_items(self):
        for dirname in _get_dirs(
            __kupfer_settings__["exclude"],
            __kupfer_settings__["min_score"],
            __kupfer_settings__["existing"],
            __kupfer_settings__["max_items"],
        ):
            yield FileLeaf(dirname)

    def _on_setting_changed(self, settings, key, value):
        if key in ("exclude", "min_score"):
            self.mark_for_update()

    def get_gicon(self):
        return icons.ComposedIconSmall("folder", "emblem-favorite")



from __future__ import annotations

import os
import pickle
import time
import typing as ty
from collections import defaultdict
from contextlib import suppress
from pathlib import Path

from kupfer import config
from kupfer.support import conspickle, pretty

if ty.TYPE_CHECKING:
    from kupfer.obj.base import Action, KupferObject, Leaf

__all__ = (
    "add_favorite",
    "erase_object_affinity",
    "get_correlation_bonus",
    "get_object_has_affinity",
    "get_record_score",
    "is_favorite",
    "load",
    "record_action_activations",
    "record_search_hit",
    "remove_favorite",
    "replace_favorites",
    "save",
    "set_correlation",
    "unregister",
)

_MNEMONICS_FILENAME = "mnemonics.pickle"
_MNEMONICS_KEY: ty.Final = "kupfer.bonus.mnemonics"
_CORRELATION_KEY: ty.Final = "kupfer.bonus.correlation"
_CORRELATION2_KEY: ty.Final = "kupfer.bonus.correlation2"
_ACTIVATIONS_KEY: ty.Final = "kupfer.bonus.activations"


# limit number of stored items in Registry
_MAX_MNEMONICS_COUNT: ty.Final[int] = 500
_MAX_CORRELATIONS_COUNT: ty.Final[int] = 100
_MAX_CORRELATION_ACTIONS: ty.Final[int] = 10


## this is a harmless default
_DEFAULT_ACTIONS: ty.Final = {
    "<kupfer.obj.apps.AppLeaf gnome-terminal>": "<kupfer.obj.apps.LaunchAgain>",
    "<kupfer.obj.apps.AppLeaf xfce4-terminal>": "<kupfer.obj.apps.LaunchAgain>",
}

# skip this actions when recording correlations
_IGNORED_ACTIONS: ty.Final = {
    "<kupfer.plugin.core.Rescan>",
    "<kupfer.plugin.applications.SetDefaultApplication>",
    "<kupfer.plugin.core.debug.DebugInfo>",
}

## Favorites is set of favorites (repr(obj))
_FAVORITES: ty.Final[set[str]] = set()
## _PLUG_FAVS are favorites by plugin; to use must be merged to _FAVORITES
_PLUG_FAVS: ty.Final[dict[str, list[str]]] = {}


class Mnemonics:
    """Class to describe a collection of mnemonics as well as the total count."""

    __slots__ = ("count", "last_ts_used", "mnemonics")

    def __init__(self) -> None:
        # map word -> number of uses
        self.mnemonics: defaultdict[str, int] = defaultdict(int)
        # total number of uses
        self.count: int = 0
        # last update timestamp
        self.last_ts_used: int = 0

    def __repr__(self) -> str:
        mnm = ", ".join(f"{m}:{c}" for m, c in self.mnemonics.items())
        last_used = (time.time() - self.last_ts_used) // 86400
        return (
            f"<{self.__class__.__name__} cnt={self.count} mnm={mnm}"
            f" ts={self.last_ts_used} ({last_used}d ago)>"
        )

    def increment(self, mnemonic: str | None = None) -> None:
        if mnemonic:
            self.mnemonics[mnemonic] += 1

        self.count += 1
        self.last_ts_used = int(time.time())

    def decrement(self) -> None:
        """Decrement total count and the least mnemonic"""
        if self.mnemonics:
            key = min(self.mnemonics.keys(), key=lambda k: self.mnemonics[k])
            if (mcount := self.mnemonics[key]) > 1:
                self.mnemonics[key] = mcount - 1
            else:
                self.mnemonics.pop(key)

        self.count = max(self.count - 1, 0)

    def score_for_key(self, key: str) -> int:
        if not key:
            # if no key, score depend on number of mnemonic usage (count)
            return 50 - 50 // (self.count + 1)

        closescr = sum(
            val for m, val in self.mnemonics.items() if m.startswith(key)
        )
        exact = self.mnemonics.get(key, 0)
        return 80 - int(50.0 / (closescr + 1) + 30.0 / (exact + 1))

    def prune(self) -> None:
        keys_to_del = [k for k, c in self.mnemonics.items() if c == 0]
        for k in keys_to_del:
            del self.mnemonics[k]

    def __getstate__(self) -> dict[str, ty.Any]:
        return {
            "count": self.count,
            "mnemonics": self.mnemonics,
            "last_ts_used": self.last_ts_used,
        }

    def __setstate__(self, state: dict[str, ty.Any]) -> None:
        self.count = state.get("count", 0)
        self.last_ts_used = state.get("last_ts_used", 0)
        self.mnemonics = defaultdict(int, state.get("mnemonics", {}))


class Correlation:
    """Correlation map leaf into actions. There should be no more than couple
    of actions."""

    __slots__ = ("actions", "last_ts_used", "permanent_action")

    def __init__(self, /, *actions: str) -> None:
        # list of actions sorted by usage (first item in list is last used
        # action)
        self.actions: list[str] = list(actions)
        # if user set this correlation this is permanent action and have extra
        # bonus
        self.permanent_action: str | None = None
        # last used timestamp
        self.last_ts_used: int = 0

    def __repr__(self) -> str:
        last_used = (time.time() - self.last_ts_used) // 86400
        return (
            f"<{self.__class__.__name__} actions={self.actions} "
            f"permanent_action={self.permanent_action} "
            f"ts={self.last_ts_used} ({last_used}d ago)>"
        )

    def score_for_action(self, action: str) -> int:
        """Score correlation for action; permanent_action give const bonus,
        for other actions - score depend on last use of given action."""
        if action == self.permanent_action:
            return 50

        for idx, a in enumerate(self.actions):
            if a == action:
                return max(1, 20 - idx)

        return 0

    def update(self, key: str) -> None:
        self.last_ts_used = int(time.time())

        if key == self.permanent_action:
            # do no update actions when action is set as permanent
            return

        with suppress(ValueError):
            self.actions.remove(key)

        self.actions.insert(0, key)

        if len(self.actions) > _MAX_CORRELATION_ACTIONS:
            self.actions.pop(-1)

    def update_permanent_action(self, key: str) -> None:
        self.permanent_action = key
        self.last_ts_used = int(time.time())


class Learning:
    @classmethod
    def unpickle_register(cls, pickle_file: str) -> dict[str, ty.Any] | None:
        try:
            pfile = Path(pickle_file).read_bytes()
            data = conspickle.ConservativeUnpickler.loads(pfile)
            assert isinstance(data, dict), "Stored object not a dict"
            pretty.print_debug(__name__, f"Reading from {pickle_file}")
            return data
        except OSError:
            pass
        except (pickle.PickleError, Exception) as exc:
            pretty.print_error(__name__, f"Error loading {pickle_file}: {exc}")

        return None

    @classmethod
    def pickle_register(cls, reg: dict[str, ty.Any], pickle_file: str) -> bool:
        ## Write to tmp then rename over for atomicity
        tmp_pickle_file = f"{pickle_file}.{os.getpid()}"
        pretty.print_debug(__name__, f"Saving to {pickle_file}")
        Path(tmp_pickle_file).write_bytes(
            pickle.dumps(reg, pickle.HIGHEST_PROTOCOL)
        )
        os.rename(tmp_pickle_file, pickle_file)
        return True


class _Register:
    def __init__(self) -> None:
        # correlations map leaf  to action
        self.correlations: defaultdict[str, Correlation] = defaultdict(
            Correlation
        )
        # mnemonics map object into Mnemonics; each mnemonic has map list of
        # words used to activate object with number of activations.
        self.mnemonics: defaultdict[str, Mnemonics] = defaultdict(Mnemonics)

    def __bool__(self) -> bool:
        return bool(self.correlations) or bool(self.mnemonics)

    def prune(self) -> None:
        self._prune_mnemonics(_MAX_MNEMONICS_COUNT)
        self._prune_correlations(_MAX_CORRELATIONS_COUNT)
        pretty.print_debug(
            __name__,
            f"Register after prune: correlations: {len(self.correlations)}, "
            f"mnemonics: {len(self.mnemonics)}.",
        )

    def _prune_correlations(self, goalitems: int) -> None:
        """Prune  old correlations up to @goalitems. This clean register
        from items for not existing any more leaves. Keep correlation
        with user-defined actions."""

        if len(self.correlations) < goalitems:
            pretty.print_debug(
                __name__,
                "Pruned correlations register not required "
                f"({len(self.correlations)} / {goalitems})",
            )
            return

        items: list[tuple[int, str]] = sorted(
            (c.last_ts_used, leaf)
            for leaf, c in self.correlations.items()
            if not c.permanent_action
        )

        to_del = items[: len(items) - goalitems]
        for _ts, leaf in to_del:
            del self.correlations[leaf]

        pretty.print_debug(
            __name__,
            "Pruned correlation from register; deleted: {len(to_del)})",
        )

    def _prune_mnemonics_unused(self) -> None:
        """Prune unused mnemonics: remove mnemonics keys with 0 value, remove
        whole mnemonic with count == 0. This clean mnemonics create by bug
        in code...
        """
        for m in self.mnemonics.values():
            m.prune()

        to_del = [k for k, m in self.mnemonics.items() if not m.count]
        for leaf in to_del:
            del self.mnemonics[leaf]

        pretty.print_debug(
            __name__,
            f"Pruned unused mnemonics from register; deleted: {len(to_del)}",
        )

    def _prune_mnemonics_decrement(self, goalitems: int) -> None:
        """Decrement mnemonics sorted by last use time; if mnemonic count == 0
        remove it."""

        # get all items sorted by last used time
        items: list[tuple[int, str, Mnemonics]] = sorted(
            (mne.last_ts_used, leaf, mne)
            for leaf, mne in self.mnemonics.items()
        )
        to_del = []
        to_del_cnt = len(items) - goalitems

        for _ts, leaf, mne in items:
            mne.decrement()
            if not mne.count:
                to_del.append(leaf)
                to_del_cnt -= 1
                if not to_del_cnt:
                    break

        for leaf in to_del:
            del self.mnemonics[leaf]

        pretty.print_debug(
            __name__,
            f"Pruned mnemonics from register; deleted: {len(to_del)}",
        )

    def _prune_mnemonics_least_used(self, goalitems: int) -> None:
        # get all items sorted by last used time
        items: list[tuple[int, str]] = sorted(
            (mne.count, leaf) for leaf, mne in self.mnemonics.items()
        )

        to_del = items[: len(self.mnemonics) - goalitems]
        for _cnt, leaf in to_del:
            del self.mnemonics[leaf]

        pretty.print_debug(
            __name__,
            f"Pruned least used mnemonics from register; deleted: {len(to_del)}",
        )

    def _prune_mnemonics(self, goalitems: int = 50) -> None:
        """Try to reduce number of mnemonic to `goalitems`.

        Always delete unused/broken mnemonics (count=0)

        To reach `goalitems`:
        1. from oldest mnemonics - decrement count & mnemonics usage; delete
           mnemonics with count == 0
        2. delete least used
        """
        self._prune_mnemonics_unused()

        if len(self.mnemonics) < goalitems:
            pretty.print_debug(
                __name__,
                "Pruned mnemonics register not required "
                f"({len(self.mnemonics)} / {goalitems})",
            )
            return

        if len(self.mnemonics) > goalitems:
            self._prune_mnemonics_decrement(goalitems)

        if len(self.mnemonics) > goalitems:
            self._prune_mnemonics_least_used(goalitems)


_REGISTER = _Register()


def record_search_hit(obj: ty.Any, key: str | None = None) -> None:
    """Record that KupferObject @obj was used, with the optional
    search term @key recording.
    When key is None - skip registration (this is only valid when action is
    performed by accelerator)."""
    if key is not None:
        _REGISTER.mnemonics[repr(obj)].increment(key)


def get_record_score(obj: ty.Any, key: str = "") -> int:
    """Get total score for KupferObject @obj, bonus score is given for @key
    matches"""
    name = repr(obj)
    fav = 7 if name in _FAVORITES else 0

    if mns := _REGISTER.mnemonics.get(name):
        return fav + mns.score_for_key(key)

    return fav


def get_correlation_bonus(action: Action, for_leaf: Leaf | None) -> int:
    """Get the bonus rank for @obj when used with @for_leaf."""
    # favorites
    repr_action = repr(action)
    repr_leaf = repr(for_leaf)

    # get bonus for exact match leaf -> action or user defined
    # correlation for leaf and action
    if (c := _REGISTER.correlations.get(repr_leaf)) and (
        s := c.score_for_action(repr_action)
    ):
        return s

    # get bonus for match type of leaf -> action
    if (c := _REGISTER.correlations.get(repr(type(for_leaf)))) and (
        s := c.score_for_action(repr_action)
    ):
        return s // 2

    return 0


def set_correlation(action: Action, for_leaf: Leaf) -> None:
    """Register @obj to get a bonus when used with @for_leaf."""
    repr_for_leaf = repr(for_leaf)
    repr_action = repr(action)
    _REGISTER.correlations[repr_for_leaf].update_permanent_action(repr_action)


def record_action_activations(action: Action, for_leaf: Leaf) -> None:
    """Record action activation for leaf that boost this action in next search.
    Also registered is object class so using this action for similar object
    also get some (smaller) bonus."""

    repr_action = repr(action)

    if repr_action in _IGNORED_ACTIONS:
        return

    _REGISTER.correlations[repr(for_leaf)].update(repr_action)
    _REGISTER.correlations[repr(type(for_leaf))].update(repr_action)


def get_object_has_affinity(obj: Leaf) -> bool:
    """Return if @obj has any positive score in the register."""
    robj = repr(obj)
    return bool(robj in _REGISTER.mnemonics or robj in _REGISTER.correlations)


def erase_object_affinity(obj: Leaf) -> None:
    """Remove all track of affinity for @obj."""
    robj = repr(obj)
    _REGISTER.mnemonics.pop(robj, None)
    _REGISTER.correlations.pop(robj, None)


def _upgrade_and_fill_register(reg: dict[str, ty.Any]) -> None:
    """Update correlation stored in old format."""
    _REGISTER.correlations.clear()
    _REGISTER.mnemonics.clear()

    # old format - activations
    if acts := reg.pop(_ACTIVATIONS_KEY, None):
        _REGISTER.correlations.update(
            {
                lr: Correlation(act)
                for lr, act in acts.items()
                if isinstance(act, str)
            }
        )

    # old format - correlations
    corrs = reg.pop(_CORRELATION_KEY, None) or _DEFAULT_ACTIONS
    # check if this is old file format (values are strings); ignore other
    # types
    if isinstance(next(iter(corrs.values())), str):
        for lr, act in corrs.items():
            _REGISTER.correlations[lr].update_permanent_action(act)

    # rest in `reg` are mnemonics
    _REGISTER.mnemonics.update(reg)


def load() -> None:
    """Load learning database."""
    if (filepath := config.get_config_file(_MNEMONICS_FILENAME)) and (
        reg := Learning.unpickle_register(filepath)
    ):
        mns = reg.get(_MNEMONICS_KEY)
        if mns:
            # mnemonics are default dict
            assert isinstance(mns, defaultdict)
            _REGISTER.mnemonics = mns

        corrs = reg.get(_CORRELATION2_KEY)
        if corrs:
            assert isinstance(corrs, defaultdict)
            _REGISTER.correlations = corrs

        if mns or corrs or not reg:
            # we get mnemonics and correlations in new format so we finish
            return

        # old file format - upgrade
        pretty.print_info(__name__, "upgrading learning registry")
        _upgrade_and_fill_register(reg)

    else:
        # load defaults
        for leaf, act in _DEFAULT_ACTIONS.items():
            _REGISTER.correlations[leaf].update_permanent_action(act)


def save() -> None:
    """Save the learning record."""

    if not _REGISTER:
        pretty.print_debug(__name__, "Not writing empty register")
        return

    _REGISTER.prune()

    filepath = config.save_config_file(_MNEMONICS_FILENAME)
    assert filepath

    reg: dict[str, ty.Any] = {
        _MNEMONICS_KEY: _REGISTER.mnemonics,
        _CORRELATION2_KEY: _REGISTER.correlations,
    }
    Learning.pickle_register(reg, filepath)


def _rebuild_favorites():
    """Build _FAVORITES set from _PLUG_FAVS buckets."""
    _FAVORITES.clear()
    for favs in _PLUG_FAVS.values():
        _FAVORITES.update(favs)


def add_favorite(plugin_id: str, *objs: KupferObject) -> None:
    """Add favorites `objs` to `plugin_id` bucket."""
    fset = _PLUG_FAVS.get(plugin_id)
    if fset:
        # filter and add new obj
        nfset = [obj for obj in map(repr, objs) if obj not in fset]
        fset.extend(nfset)
    else:
        nfset = _PLUG_FAVS[plugin_id] = list(map(repr, objs))

    _FAVORITES.update(nfset)


def replace_favorites(plugin_id: str, *objs: KupferObject) -> None:
    """Replace favorites in `plugin_id` bucket. If no `objs` remove all
    favorites for `plugin_id`."""
    if not objs:
        if plugin_id in _PLUG_FAVS:
            del _PLUG_FAVS[plugin_id]
            _rebuild_favorites()

        return

    fset = _PLUG_FAVS[plugin_id] = list(map(repr, objs))
    _FAVORITES.update(fset)


def remove_favorite(plugin_id: str, obj: KupferObject) -> None:
    """Remove `obj` from favorites in `plugin_id` bucket."""
    if fset := _PLUG_FAVS.get(plugin_id):
        try:
            fset.remove(repr(obj))
        except ValueError:
            return

        if fset:
            _PLUG_FAVS[plugin_id] = fset
        else:
            del _PLUG_FAVS[plugin_id]

        _rebuild_favorites()


def is_favorite(obj: KupferObject) -> bool:
    return repr(obj) in _FAVORITES


def unregister(obj):
    _REGISTER.mnemonics.pop(obj, None)
